import string
import random

"""
File for utility functions used in testing
"""


def random_string_of_length(length):
    """
    Generates a random string of ASCII letters of a fixed length.

    Args:
        length (int): fixed length desired

    Returns:
        string: random string generated
    """
    return "".join(random.choices(string.ascii_letters, k=length))



import pytest
from tests.mocks import MockOktaClient
import okta.models as models
from http import HTTPStatus
from okta.errors.okta_api_error import OktaAPIError


class TestTrustedOriginsResource:
    """
    Integration Tests for the Trusted Origins Resource
    """
    SDK_PREFIX = "python_sdk"

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_create_get_origin(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Trusted Origin
        TO_NAME = f"{TestTrustedOriginsResource.SDK_PREFIX}_test_TO"
        TO_ORIGIN = "http://example.com"
        trusted_origin_model = models.TrustedOrigin({
            "name": TO_NAME,
            "origin": TO_ORIGIN,
            "scopes": [
                models.Scope({
                    "type": models.ScopeType.CORS
                }),
                models.Scope({
                    "type": models.ScopeType.REDIRECT
                }),
            ]
        })

        created_trusted_origin, _, err = await \
            client.create_origin(trusted_origin_model)
        assert err is None
        assert isinstance(created_trusted_origin, models.TrustedOrigin)

        # Retrieve
        retrieved_origin, _, err = await \
            client.get_origin(created_trusted_origin.id)
        assert err is None
        assert isinstance(retrieved_origin, models.TrustedOrigin)
        assert retrieved_origin.name == created_trusted_origin.name
        assert len(retrieved_origin.scopes) == 2

        # Delete
        _, err = await client.delete_origin(created_trusted_origin.id)

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_list_origins(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Trusted Origin
        TO_NAME = f"{TestTrustedOriginsResource.SDK_PREFIX}_test_TO"
        TO_ORIGIN = "http://example.com"
        trusted_origin_model = models.TrustedOrigin({
            "name": TO_NAME,
            "origin": TO_ORIGIN,
            "scopes": [
                models.Scope({
                    "type": models.ScopeType.CORS
                }),
                models.Scope({
                    "type": models.ScopeType.REDIRECT
                }),
            ]
        })

        created_trusted_origin, _, err = await \
            client.create_origin(trusted_origin_model)
        assert err is None
        assert isinstance(created_trusted_origin, models.TrustedOrigin)

        # List
        trusted_origins, _, err = await client.list_origins()
        assert err is None
        assert isinstance(trusted_origins, list)
        assert len(trusted_origins) > 0
        assert isinstance(trusted_origins[0], models.TrustedOrigin)
        assert next((to for to in trusted_origins
                     if to.name == created_trusted_origin.name), None) \
            is not None

        # Delete
        _, err = await client.delete_origin(created_trusted_origin.id)

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_delete_origin(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Trusted Origin
        TO_NAME = f"{TestTrustedOriginsResource.SDK_PREFIX}_test_TO"
        TO_ORIGIN = "http://example.com"
        trusted_origin_model = models.TrustedOrigin({
            "name": TO_NAME,
            "origin": TO_ORIGIN,
            "scopes": [
                models.Scope({
                    "type": models.ScopeType.CORS
                }),
                models.Scope({
                    "type": models.ScopeType.REDIRECT
                }),
            ]
        })

        created_trusted_origin, _, err = await \
            client.create_origin(trusted_origin_model)
        assert err is None
        assert isinstance(created_trusted_origin, models.TrustedOrigin)

        # Retrieve
        retrieved_origin, _, err = await \
            client.get_origin(created_trusted_origin.id)
        assert err is None
        assert isinstance(retrieved_origin, models.TrustedOrigin)
        assert retrieved_origin.name == created_trusted_origin.name
        assert len(retrieved_origin.scopes) == 2

        # Delete
        _, err = await client.delete_origin(created_trusted_origin.id)

        # Retrieve to validate
        retrieved_origin, resp, err = await \
            client.get_origin(created_trusted_origin.id)
        assert err is not None
        assert isinstance(err, OktaAPIError)
        assert resp.get_status() == HTTPStatus.NOT_FOUND
        assert retrieved_origin is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_update_origin(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Trusted Origin
        TO_NAME = f"{TestTrustedOriginsResource.SDK_PREFIX}_test_TO"
        TO_ORIGIN = "http://example.com"
        trusted_origin_model = models.TrustedOrigin({
            "name": TO_NAME,
            "origin": TO_ORIGIN,
            "scopes": [
                models.Scope({
                    "type": models.ScopeType.CORS
                }),
                models.Scope({
                    "type": models.ScopeType.REDIRECT
                }),
            ]
        })

        created_trusted_origin, _, err = await \
            client.create_origin(trusted_origin_model)
        assert err is None
        assert isinstance(created_trusted_origin, models.TrustedOrigin)

        # Retrieve
        retrieved_origin, _, err = await \
            client.get_origin(created_trusted_origin.id)
        assert err is None
        assert isinstance(retrieved_origin, models.TrustedOrigin)
        assert retrieved_origin.name == created_trusted_origin.name
        assert len(retrieved_origin.scopes) == 2

        # Update
        updated_trusted_origin_model = models.TrustedOrigin({
            "name": TO_NAME + "_updated",
            "origin": TO_ORIGIN,
            "scopes": [
                models.Scope({
                    "type": models.ScopeType.CORS
                }),
                models.Scope({
                    "type": models.ScopeType.REDIRECT
                }),
            ]
        })
        updated_origin, _, err = await \
            client.update_origin(created_trusted_origin.id,
                                 updated_trusted_origin_model)
        assert err is None
        assert isinstance(updated_origin, models.TrustedOrigin)
        assert updated_origin.id == created_trusted_origin.id
        assert updated_origin.name == updated_trusted_origin_model.name

        # Retrieve to validate
        retrieved_origin, resp, err = await \
            client.get_origin(created_trusted_origin.id)
        assert retrieved_origin.id == created_trusted_origin.id
        assert retrieved_origin.name == updated_origin.name

        # Delete
        _, err = await client.delete_origin(created_trusted_origin.id)

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_activate_deactivate_origin(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Trusted Origin
        TO_NAME = f"{TestTrustedOriginsResource.SDK_PREFIX}_test_TO"
        TO_ORIGIN = "http://example.com"
        trusted_origin_model = models.TrustedOrigin({
            "name": TO_NAME,
            "origin": TO_ORIGIN,
            "scopes": [
                models.Scope({
                    "type": models.ScopeType.CORS
                }),
                models.Scope({
                    "type": models.ScopeType.REDIRECT
                }),
            ]
        })

        created_trusted_origin, _, err = await \
            client.create_origin(trusted_origin_model)
        assert err is None
        assert isinstance(created_trusted_origin, models.TrustedOrigin)
        assert created_trusted_origin.status == "ACTIVE"

        # Deactivate
        deactivated_origin, _, err = await \
            client.deactivate_origin(created_trusted_origin.id)
        assert err is None
        assert deactivated_origin.status == "INACTIVE"

        # Retrieve to validate
        retrieved_origin, resp, err = await \
            client.get_origin(created_trusted_origin.id)
        assert retrieved_origin.id == created_trusted_origin.id
        assert retrieved_origin.status == "INACTIVE"

        # Reactivate
        reactivated_origin, _, err = await \
            client.activate_origin(created_trusted_origin.id)
        assert err is None
        assert reactivated_origin.status == "ACTIVE"

        # Retrieve to validate
        retrieved_origin, resp, err = await \
            client.get_origin(created_trusted_origin.id)
        assert retrieved_origin.id == created_trusted_origin.id
        assert retrieved_origin.status == "ACTIVE"

        # Delete
        _, err = await client.delete_origin(created_trusted_origin.id)



import pytest
from tests.mocks import MockOktaClient
import okta.models as models
from http import HTTPStatus
from okta.errors.okta_api_error import OktaAPIError


class TestUserTypesResource:
    """
    Integration Tests for the User Types Resource
    """
    SDK_PREFIX = "python_sdk"

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_create_user_type(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create object
        TEST_DESC = F"{TestUserTypesResource.SDK_PREFIX} description"
        TEST_DISPLAY_NAME = F"{TestUserTypesResource.SDK_PREFIX} display name"
        TEST_NAME = F"{TestUserTypesResource.SDK_PREFIX}_aAaewb"
        # ^ without VCR testing, can generate random TEST_NAME suffixes

        user_type_obj = models.UserType({
            "description": TEST_DESC,
            "displayName": TEST_DISPLAY_NAME,
            "name": TEST_NAME
        })

        created, _, err = await client.create_user_type(user_type_obj)
        assert err is None
        assert created.id is not None
        assert created.description == TEST_DESC
        assert created.display_name == TEST_DISPLAY_NAME
        assert created.name == TEST_NAME

        _, err = await client.delete_user_type(created.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_get_user_type(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create object
        TEST_DESC = F"{TestUserTypesResource.SDK_PREFIX} description"
        TEST_DISPLAY_NAME = F"{TestUserTypesResource.SDK_PREFIX} display name"
        TEST_NAME = F"{TestUserTypesResource.SDK_PREFIX}_akIKx"
        # ^ without VCR testing, can generate random TEST_NAME suffixes

        user_type_obj = models.UserType({
            "description": TEST_DESC,
            "displayName": TEST_DISPLAY_NAME,
            "name": TEST_NAME
        })

        created, _, err = await client.create_user_type(user_type_obj)
        assert err is None
        assert created.id is not None

        # Retrieve
        found, _, err = await client.get_user_type(created.id)
        assert err is None
        assert found.id == created.id
        assert found.description == TEST_DESC
        assert found.display_name == TEST_DISPLAY_NAME
        assert found.name == TEST_NAME

        _, err = await client.delete_user_type(created.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_update_user_type(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create object
        TEST_DESC = F"{TestUserTypesResource.SDK_PREFIX} description"
        TEST_DISPLAY_NAME = F"{TestUserTypesResource.SDK_PREFIX} display name"
        TEST_NAME = F"{TestUserTypesResource.SDK_PREFIX}_JWXbL"
        # ^ without VCR testing, can generate random TEST_NAME suffixes

        user_type_obj = models.UserType({
            "description": TEST_DESC,
            "displayName": TEST_DISPLAY_NAME,
            "name": TEST_NAME
        })

        created, _, err = await client.create_user_type(user_type_obj)
        assert err is None
        assert created.id is not None

        # Update
        UPDATED_TEST_DESC = f"{TEST_DESC}-updated"
        UPDATED_TEST_DISPLAY_NAME = f"{TEST_DISPLAY_NAME}-updated"
        updated_obj = models.UserType({
            "description": UPDATED_TEST_DESC,
            "displayName": UPDATED_TEST_DISPLAY_NAME,
        })
        updated, _, err = await client.update_user_type(created.id,
                                                        updated_obj)
        assert err is None
        assert updated.id == created.id
        assert updated.description == UPDATED_TEST_DESC
        assert updated.display_name == UPDATED_TEST_DISPLAY_NAME

        # Retrieve
        found, _, err = await client.get_user_type(created.id)
        assert err is None
        assert found.id == created.id
        assert found.description == UPDATED_TEST_DESC
        assert found.display_name == UPDATED_TEST_DISPLAY_NAME

        _, err = await client.delete_user_type(created.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_replace_user_type(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create object
        TEST_DESC = F"{TestUserTypesResource.SDK_PREFIX} description"
        TEST_DISPLAY_NAME = F"{TestUserTypesResource.SDK_PREFIX} display name"
        TEST_NAME = F"{TestUserTypesResource.SDK_PREFIX}_XfloX"
        # ^ without VCR testing, can generate random TEST_NAME suffixes

        user_type_obj = models.UserType({
            "description": TEST_DESC,
            "displayName": TEST_DISPLAY_NAME,
            "name": TEST_NAME
        })

        created, _, err = await client.create_user_type(user_type_obj)
        assert err is None
        assert created.id is not None

        # Replace
        REPLACED_TEST_DESC = f"{TEST_DESC}-replaced"
        REPLACED_TEST_DISPLAY_NAME = f"{TEST_DISPLAY_NAME}-replaced"
        replaced_obj = models.UserType({
            "description": REPLACED_TEST_DESC,
            "displayName": REPLACED_TEST_DISPLAY_NAME,
            "name": TEST_NAME
        })
        replaced, _, err = await client.replace_user_type(created.id,
                                                          replaced_obj)
        assert err is None
        assert replaced.id == created.id
        assert replaced.description == REPLACED_TEST_DESC
        assert replaced.display_name == REPLACED_TEST_DISPLAY_NAME
        assert replaced.name == TEST_NAME

        # Retrieve
        found, _, err = await client.get_user_type(created.id)
        assert err is None
        assert found.id == created.id
        assert found.description == REPLACED_TEST_DESC
        assert found.display_name == REPLACED_TEST_DISPLAY_NAME
        assert found.name == TEST_NAME

        _, err = await client.delete_user_type(created.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_delete_user_type(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create object
        TEST_DESC = F"{TestUserTypesResource.SDK_PREFIX} description"
        TEST_DISPLAY_NAME = F"{TestUserTypesResource.SDK_PREFIX} display name"
        TEST_NAME = F"{TestUserTypesResource.SDK_PREFIX}_zEVAT"
        # ^ without VCR testing, can generate random TEST_NAME suffixes

        user_type_obj = models.UserType({
            "description": TEST_DESC,
            "displayName": TEST_DISPLAY_NAME,
            "name": TEST_NAME
        })

        created, _, err = await client.create_user_type(user_type_obj)
        assert err is None
        assert created.id is not None

        # Retrieve
        found, _, err = await client.get_user_type(created.id)
        assert err is None
        assert found.id == created.id
        assert found.description == TEST_DESC
        assert found.display_name == TEST_DISPLAY_NAME
        assert found.name == TEST_NAME

        # Delete
        _, err = await client.delete_user_type(created.id)
        assert err is None

        # Retrieve
        found, resp, err = await client.get_user_type(created.id)
        assert err is not None
        assert isinstance(err, OktaAPIError)
        assert resp.get_status() == HTTPStatus.NOT_FOUND

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_list_user_types(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create object
        TEST_DESC = F"{TestUserTypesResource.SDK_PREFIX} description"
        TEST_DISPLAY_NAME = F"{TestUserTypesResource.SDK_PREFIX} display name"
        TEST_NAME = F"{TestUserTypesResource.SDK_PREFIX}_cRuLD"
        TEST_DESC_2 = F"{TestUserTypesResource.SDK_PREFIX} description-2"
        TEST_DISPLAY_NAME_2 = F"{TestUserTypesResource.SDK_PREFIX} dn-2"
        TEST_NAME_2 = F"{TestUserTypesResource.SDK_PREFIX}_cRuLD_2"
        # ^ without VCR testing, can generate random TEST_NAME suffixes

        user_type_obj = models.UserType({
            "description": TEST_DESC,
            "displayName": TEST_DISPLAY_NAME,
            "name": TEST_NAME
        })
        user_type_obj_2 = models.UserType({
            "description": TEST_DESC_2,
            "displayName": TEST_DISPLAY_NAME_2,
            "name": TEST_NAME_2
        })

        ut1, _, err = await client.create_user_type(user_type_obj)
        assert err is None
        assert ut1.id is not None
        ut2, _, err = await client.create_user_type(user_type_obj_2)
        assert err is None
        assert ut2.id is not None

        # List
        user_types, _, err = await client.list_user_types()
        assert err is None
        assert len(user_types) > 0
        assert next((ut for ut in user_types if ut.id == ut1.id))
        assert next((ut for ut in user_types if ut.id == ut2.id))

        _, err = await client.delete_user_type(ut1.id)
        assert err is None
        _, err = await client.delete_user_type(ut2.id)
        assert err is None



import pytest
from tests.mocks import MockOktaClient
import okta.models as models
from okta.errors.okta_api_error import OktaAPIError


class TestApplicationsResource:
    """
    Integration Tests for the Applications Resource
    """

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_create_bookmark_app(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Bookmark Application Object
        APP_URL = "https://example.com/bookmark.htm"
        APP_LABEL = "AddBookmarkApp"
        app_settings_app = models.BookmarkApplicationSettingsApplication({
            "requestIntegration": False,
            "url": APP_URL
        })
        app_settings = models.BookmarkApplicationSettings({
            "app": app_settings_app
        })
        bookmark_app_obj = models.BookmarkApplication({
            "label": APP_LABEL,
            "settings": app_settings
        })

        # Create App in org
        app, _, err = await client.create_application(bookmark_app_obj)
        assert err is None
        assert isinstance(app, models.Application)
        assert isinstance(app, models.BookmarkApplication)

        # Get app and verify details
        found_app, _, err = await client.get_application(app.id)
        assert err is None
        assert found_app.label == APP_LABEL
        assert found_app.sign_on_mode == models.ApplicationSignOnMode.BOOKMARK
        assert found_app.settings.app.request_integration is False
        assert found_app.settings.app.url == APP_URL

        # Deactivate & Delete created app
        _, err = await client.deactivate_application(app.id)
        assert err is None
        _, err = await client.delete_application(app.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_create_basic_auth_app(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create BasicAuth Application Object
        APP_AUTH_URL = "https://example.com/auth.html"
        APP_URL = "https://example.com/auth.html"
        APP_LABEL = "AddBasicAuthApp"
        app_settings_app = models.BasicApplicationSettingsApplication({
            "authUrl": APP_AUTH_URL,
            "url": APP_URL
        })
        app_settings = models.BasicApplicationSettings({
            "app": app_settings_app
        })
        basic_auth_app_obj = models.BasicAuthApplication({
            "label": APP_LABEL,
            "settings": app_settings
        })

        # Create App in org
        app, _, err = await client.create_application(basic_auth_app_obj)
        assert err is None
        assert isinstance(app, models.Application)
        assert isinstance(app, models.BasicAuthApplication)

        # Get app and verify details
        found_app, _, err = await client.get_application(app.id)
        assert err is None
        assert found_app.label == APP_LABEL
        assert found_app.sign_on_mode == \
            models.ApplicationSignOnMode.BASIC_AUTH
        assert found_app.settings.app.url == APP_URL
        assert found_app.settings.app.auth_url == APP_AUTH_URL

        # Deactivate & Delete created app
        _, err = await client.deactivate_application(app.id)
        assert err is None
        _, err = await client.delete_application(app.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_create_SWA_app(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Application object and Application in Org
        APP_LABEL = "Test App"
        BUTTON_FIELD = "btn-login"
        PASSWORD_FIELD = "txt-box-password"
        USERNAME_FIELD = "txt-box-username"
        URL = "https://example.com/login.html"
        LOGIN_URL_REGEX = f"^{URL}$"

        swa_app_settings_app = models.SwaApplicationSettingsApplication({
            "buttonField": BUTTON_FIELD,
            "passwordField": PASSWORD_FIELD,
            "usernameField": USERNAME_FIELD,
            "url": URL,
            "loginUrlRegex": LOGIN_URL_REGEX
        })

        swa_app_settings = models.SwaApplicationSettings({
            "app": swa_app_settings_app
        })

        swa_app_obj = models.SwaApplication({
            "label": APP_LABEL,
            "settings": swa_app_settings,
        })

        app, _, err = await client.create_application(swa_app_obj)
        assert err is None
        assert isinstance(app, models.Application)
        assert isinstance(app, models.SwaApplication)

        # Get app and verify details
        found_app, _, err = await client.get_application(app.id)
        assert err is None
        assert found_app.label == APP_LABEL
        assert found_app.sign_on_mode == \
            models.ApplicationSignOnMode.BROWSER_PLUGIN
        assert found_app.settings.app.button_field == BUTTON_FIELD
        assert found_app.settings.app.password_field == PASSWORD_FIELD
        assert found_app.settings.app.username_field == USERNAME_FIELD
        assert found_app.settings.app.url == URL
        assert found_app.settings.app.login_url_regex == LOGIN_URL_REGEX

        # Deactivate & Delete created app
        _, err = await client.deactivate_application(app.id)
        assert err is None
        _, err = await client.delete_application(app.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_create_SWA_three_app(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Application object and Application in Org
        APP_LABEL = "Test App"
        BUTTON_SELECTOR = "#btn-login"
        PASSWORD_SELECTOR = "#txt-box-password"
        USERNAME_SELECTOR = "#txt-box-username"
        TARGET_URL = "https://example.com/login.html"
        LOGIN_URL_REGEX = f"^{TARGET_URL}$"
        EXTRA_FIELD_SELECTOR = ".login"
        EXTRA_FIELD_VALUE = "SOMEVALUE"

        swa_app_settings_app =\
            models.SwaThreeFieldApplicationSettingsApplication({
                "buttonSelector": BUTTON_SELECTOR,
                "passwordSelector": PASSWORD_SELECTOR,
                "userNameSelector": USERNAME_SELECTOR,
                "targetUrl": TARGET_URL,
                "loginUrlRegex": LOGIN_URL_REGEX,
                "extraFieldSelector": EXTRA_FIELD_SELECTOR,
                "extraFieldValue": EXTRA_FIELD_VALUE
            })

        swa_app_settings = models.SwaThreeFieldApplicationSettings({
            "app": swa_app_settings_app
        })

        swa_app_obj = models.SwaThreeFieldApplication({
            "label": APP_LABEL,
            "settings": swa_app_settings,
        })

        app, _, err = await client.create_application(swa_app_obj)
        assert err is None
        assert isinstance(app, models.Application)
        assert isinstance(app, models.SwaThreeFieldApplication)

        # Get app and verify details
        found_app, _, err = await client.get_application(app.id)
        assert err is None
        assert found_app.label == APP_LABEL
        assert found_app.sign_on_mode == \
            models.ApplicationSignOnMode.BROWSER_PLUGIN
        assert found_app.settings.app.button_selector == BUTTON_SELECTOR
        assert found_app.settings.app.password_selector == PASSWORD_SELECTOR
        assert found_app.settings.app.user_name_selector == USERNAME_SELECTOR
        assert found_app.settings.app.target_url == TARGET_URL
        assert found_app.settings.app.login_url_regex == LOGIN_URL_REGEX
        assert found_app.settings.app.extra_field_selector ==\
            EXTRA_FIELD_SELECTOR
        assert found_app.settings.app.extra_field_value == EXTRA_FIELD_VALUE

        # Deactivate & Delete created app
        _, err = await client.deactivate_application(app.id)
        assert err is None
        _, err = await client.delete_application(app.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_create_secure_password_store_app(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Application object and Application in Org
        APP_LABEL = "Test App"
        PASSWORD_FIELD = "txt-box-password"
        USERNAME_FIELD = "txt-box-username"
        URL = "https://example.com/login.html"
        OPT_FIELD_1 = "param1"
        OPT_FIELD_1_VALUE = "value1"
        OPT_FIELD_2 = "param2"
        OPT_FIELD_2_VALUE = "value2"
        OPT_FIELD_3 = "param3"
        OPT_FIELD_3_VALUE = "value3"

        sps_app_settings_app =\
            models.SecurePasswordStoreApplicationSettingsApplication({
                "passwordField": PASSWORD_FIELD,
                "usernameField": USERNAME_FIELD,
                "url": URL,
                "optionalField1": OPT_FIELD_1,
                "optionalField1Value": OPT_FIELD_1_VALUE,
                "optionalField2": OPT_FIELD_2,
                "optionalField2Value": OPT_FIELD_2_VALUE,
                "optionalField3": OPT_FIELD_3,
                "optionalField3Value": OPT_FIELD_3_VALUE,
            })

        sps_app_settings = models.SecurePasswordStoreApplicationSettings({
            "app": sps_app_settings_app
        })

        sps_app_obj = models.SecurePasswordStoreApplication({
            "label": APP_LABEL,
            "settings": sps_app_settings,
        })

        app, _, err = await client.create_application(sps_app_obj)
        assert err is None
        assert isinstance(app, models.Application)
        assert isinstance(app, models.SecurePasswordStoreApplication)

        # Get app and verify details
        found_app, _, err = await client.get_application(app.id)
        assert err is None
        assert found_app.label == APP_LABEL
        assert found_app.sign_on_mode == \
            models.ApplicationSignOnMode.SECURE_PASSWORD_STORE
        assert found_app.settings.app.username_field == USERNAME_FIELD
        assert found_app.settings.app.password_field == PASSWORD_FIELD
        assert found_app.settings.app.url == URL
        assert found_app.settings.app.optional_field_1 == OPT_FIELD_1
        assert found_app.settings.app.optional_field_1_value == \
            OPT_FIELD_1_VALUE
        assert found_app.settings.app.optional_field_2 == OPT_FIELD_2
        assert found_app.settings.app.optional_field_2_value == \
            OPT_FIELD_2_VALUE
        assert found_app.settings.app.optional_field_3 == OPT_FIELD_3
        assert found_app.settings.app.optional_field_3_value == \
            OPT_FIELD_3_VALUE

        # Deactivate & Delete created app
        _, err = await client.deactivate_application(app.id)
        assert err is None
        _, err = await client.delete_application(app.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_create_open_id_connect_app(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create OIDC Application Object
        APP_LABEL = "AddOpenIdConnectApp"
        CLIENT_URI = "https://example.com/client"
        LOGO_URI = "https://example.com/assets/images/logo-new.png"
        POLICY_URI = "https://example.com/client/policy"
        GRANT_TYPES = [models.OAuthGrantType.AUTHORIZATION_CODE,
                       models.OAuthGrantType.IMPLICIT]
        APP_TYPE = models.OpenIdConnectApplicationType.NATIVE
        POST_LOGOUT_REDIRECT_URIS = ["https://example.com/postlogout",
                                     "myapp://postlogoutcallback"]
        REDIRECT_URIS = ["https://example.com/oauth2/callback",
                         "myapp://callback"]
        RESPONSE_TYPES = [models.OAuthResponseType.TOKEN,
                          models.OAuthResponseType.ID_TOKEN,
                          models.OAuthResponseType.CODE]
        TOS_URL = "https://example.com/client/tos"
        AUTO_KEY_ROTATION = True
        CLIENT_ID = "testclientid12345"
        TOKEN_ENDPOINT_AUTH_METHOD =\
            models.OAuthEndpointAuthenticationMethod.CLIENT_SECRET_POST

        app_settings_client = models.OpenIdConnectApplicationSettingsClient({
            "applicationType": APP_TYPE,
            "clientUri": CLIENT_URI,
            "grantTypes": GRANT_TYPES,
            "logoUri": LOGO_URI,
            "policyUri": POLICY_URI,
            "postLogoutRedirectUris": POST_LOGOUT_REDIRECT_URIS,
            "redirectUris": REDIRECT_URIS,
            "responseTypes": RESPONSE_TYPES,
            "tosUri": TOS_URL
        })
        app_settings = models.OpenIdConnectApplicationSettings({
            "oauthClient": app_settings_client
        })
        app_credentials = models.OAuthApplicationCredentials({
            "oauthClient": models.ApplicationCredentialsOAuthClient({
                "autoKeyRotation": AUTO_KEY_ROTATION,
                "clientId": CLIENT_ID,
                "tokenEndpointAuthMethod": TOKEN_ENDPOINT_AUTH_METHOD
            })
        })
        oidc_app_obj = models.OpenIdConnectApplication({
            "label": APP_LABEL,
            "settings": app_settings,
            "credentials": app_credentials
        })

        # Create App in org
        app, _, err = await client.create_application(oidc_app_obj)
        assert err is None
        assert isinstance(app, models.Application)
        assert isinstance(app, models.OpenIdConnectApplication)

        # Get app and verify details
        found_app, _, err = await client.get_application(app.id)
        assert err is None
        assert found_app.label == APP_LABEL
        assert found_app.sign_on_mode ==\
            models.ApplicationSignOnMode.OPENID_CONNECT
        assert found_app.settings.oauth_client.application_type == APP_TYPE
        assert found_app.settings.oauth_client.client_uri == CLIENT_URI
        assert found_app.settings.oauth_client.grant_types == GRANT_TYPES
        assert found_app.settings.oauth_client.logo_uri == LOGO_URI
        assert found_app.settings.oauth_client.post_logout_redirect_uris ==\
            POST_LOGOUT_REDIRECT_URIS
        assert found_app.settings.oauth_client.redirect_uris == REDIRECT_URIS
        assert found_app.settings.oauth_client.response_types == RESPONSE_TYPES

        # Deactivate & Delete created app
        _, err = await client.deactivate_application(app.id)
        assert err is None
        _, err = await client.delete_application(app.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_list_apps(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create OIDC Application Object
        APP_LABEL = "AddOpenIdConnectApp"
        CLIENT_URI = "https://example.com/client"
        LOGO_URI = "https://example.com/assets/images/logo-new.png"
        POLICY_URI = "https://example.com/client/policy"
        GRANT_TYPES = [models.OAuthGrantType.AUTHORIZATION_CODE,
                       models.OAuthGrantType.IMPLICIT]
        APP_TYPE = models.OpenIdConnectApplicationType.NATIVE
        POST_LOGOUT_REDIRECT_URIS = ["https://example.com/postlogout",
                                     "myapp://postlogoutcallback"]
        REDIRECT_URIS = ["https://example.com/oauth2/callback",
                         "myapp://callback"]
        RESPONSE_TYPES = [models.OAuthResponseType.TOKEN,
                          models.OAuthResponseType.ID_TOKEN,
                          models.OAuthResponseType.CODE]
        TOS_URL = "https://example.com/client/tos"
        AUTO_KEY_ROTATION = True
        CLIENT_ID = "testclientid12345"
        TOKEN_ENDPOINT_AUTH_METHOD =\
            models.OAuthEndpointAuthenticationMethod.CLIENT_SECRET_POST

        app_settings_client = models.OpenIdConnectApplicationSettingsClient({
            "applicationType": APP_TYPE,
            "clientUri": CLIENT_URI,
            "grantTypes": GRANT_TYPES,
            "logoUri": LOGO_URI,
            "policyUri": POLICY_URI,
            "postLogoutRedirectUris": POST_LOGOUT_REDIRECT_URIS,
            "redirectUris": REDIRECT_URIS,
            "responseTypes": RESPONSE_TYPES,
            "tosUri": TOS_URL
        })
        app_settings = models.OpenIdConnectApplicationSettings({
            "oauthClient": app_settings_client
        })
        app_credentials = models.OAuthApplicationCredentials({
            "oauthClient": models.ApplicationCredentialsOAuthClient({
                "autoKeyRotation": AUTO_KEY_ROTATION,
                "clientId": CLIENT_ID,
                "tokenEndpointAuthMethod": TOKEN_ENDPOINT_AUTH_METHOD
            })
        })
        oidc_app_obj = models.OpenIdConnectApplication({
            "label": APP_LABEL,
            "settings": app_settings,
            "credentials": app_credentials
        })

        # Create App in org
        app, _, err = await client.create_application(oidc_app_obj)
        assert err is None
        assert isinstance(app, models.Application)
        assert isinstance(app, models.OpenIdConnectApplication)

        # Deactivate App
        _, err = await client.deactivate_application(app.id)
        assert err is None

        # List inactive apps
        query_params_list = {"filter": "status eq \"INACTIVE\""}
        apps_list, _, err = await client.list_applications(query_params_list)
        assert err is None
        assert len(apps_list) == 1
        assert apps_list[0].id == app.id

        # Delete created app
        _, err = await client.delete_application(app.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_activate_deactivate_delete_app(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Bookmark Application Object
        APP_URL = "https://example.com/bookmark.htm"
        APP_LABEL = "AddBookmarkApp"
        app_settings_app = models.BookmarkApplicationSettingsApplication({
            "requestIntegration": False,
            "url": APP_URL
        })
        app_settings = models.BookmarkApplicationSettings({
            "app": app_settings_app
        })
        bookmark_app_obj = models.BookmarkApplication({
            "label": APP_LABEL,
            "settings": app_settings
        })

        # Create App in org
        query_params_create = {"activate": False}
        app, _, err = await client.create_application(bookmark_app_obj,
                                                      query_params_create)
        assert err is None
        assert isinstance(app, models.Application)
        assert isinstance(app, models.BookmarkApplication)
        assert app.status == "INACTIVE"

        # Activate App
        _, err = await client.activate_application(app.id)
        assert err is None

        # 2 second sleep for backend to update

        # Get app and verify details
        found_app, _, err = await client.get_application(app.id)
        assert err is None
        assert found_app.label == APP_LABEL
        assert found_app.status == "ACTIVE"

        # Try to delete active app - returns exception
        _, err = await client.delete_application(app.id)
        assert isinstance(err, OktaAPIError)

        # De-activate app
        _, err = await client.deactivate_application(app.id)
        assert err is None

        # Delete app
        _, err = await client.delete_application(app.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_assign_user_app(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Bookmark Application Object
        APP_URL = "https://example.com/bookmark.htm"
        APP_LABEL = "AddBookmarkApp"
        app_settings_app = models.BookmarkApplicationSettingsApplication({
            "requestIntegration": False,
            "url": APP_URL
        })
        app_settings = models.BookmarkApplicationSettings({
            "app": app_settings_app
        })
        bookmark_app_obj = models.BookmarkApplication({
            "label": APP_LABEL,
            "settings": app_settings
        })

        # Create App in org
        app, _, err = await client.create_application(bookmark_app_obj)
        assert err is None
        assert isinstance(app, models.Application)
        assert isinstance(app, models.BookmarkApplication)

        # Create Password
        password = models.PasswordCredential({
            "value": "Password150kta"
        })

        # Create User Credentials
        user_creds = models.UserCredentials({
            "password": password
        })

        # Create User Profile and CreateUser Request
        user_profile = models.UserProfile()
        user_profile.first_name = "John"
        user_profile.last_name = "Doe-Get"
        user_profile.email = "John.Doe-Get@example.com"
        user_profile.login = "John.Doe-Get@example.com"

        create_user_req = models.CreateUserRequest({
            "credentials": user_creds,
            "profile": user_profile
        })

        query_params_create = {"activate": True}

        # Create User
        user, resp, err = await client.create_user(
            create_user_req, query_params_create)
        assert err is None

        # Create app user
        app_user_credentials_password = models.AppUserPasswordCredential({
            "value": "Password150kta"
        })
        app_user_credentials = models.AppUserCredentials({
            "password": app_user_credentials_password,
            "userName": user.profile.email
        })
        app_user = models.AppUser({
            "credentials": app_user_credentials,
            "id": user.id
        })

        # Assign
        created_app_user, _, err = await\
            client.assign_user_to_application(app.id, app_user)
        assert err is None
        assert isinstance(created_app_user, models.AppUser)
        assert created_app_user.scope == "USER"
        assert created_app_user.status == "ACTIVE"
        assert created_app_user.sync_state == "DISABLED"

        # Deactivate and Delete User
        _, err = await client.deactivate_user(user.id)
        assert err is None
        _, err = await client.deactivate_or_delete_user(user.id)
        assert err is None

        # Deactivate & Delete created app
        _, err = await client.deactivate_application(app.id)
        assert err is None
        _, err = await client.delete_application(app.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_get_list_assign_users_app(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Bookmark Application Object
        APP_URL = "https://example.com/bookmark.htm"
        APP_LABEL = "AddBookmarkApp"
        app_settings_app = models.BookmarkApplicationSettingsApplication({
            "requestIntegration": False,
            "url": APP_URL
        })
        app_settings = models.BookmarkApplicationSettings({
            "app": app_settings_app
        })
        bookmark_app_obj = models.BookmarkApplication({
            "label": APP_LABEL,
            "settings": app_settings
        })

        # Create App in org
        app, _, err = await client.create_application(bookmark_app_obj)
        assert err is None
        assert isinstance(app, models.Application)
        assert isinstance(app, models.BookmarkApplication)

        # Create Password
        password = models.PasswordCredential({
            "value": "Password150kta"
        })

        # Create User Credentials
        user_creds = models.UserCredentials({
            "password": password
        })

        # Create User Profile and CreateUser Request
        user_profile = models.UserProfile()
        user_profile.first_name = "John"
        user_profile.last_name = "Doe-Get"
        user_profile.email = "John.Doe-Get@example.com"
        user_profile.login = "John.Doe-Get@example.com"

        user_profile_2 = models.UserProfile()
        user_profile_2.first_name = "John"
        user_profile_2.last_name = "Doe-List"
        user_profile_2.email = "John.Doe-List@example.com"
        user_profile_2.login = "John.Doe-List@example.com"

        create_user_req = models.CreateUserRequest({
            "credentials": user_creds,
            "profile": user_profile
        })
        create_user_req_2 = models.CreateUserRequest({
            "credentials": user_creds,
            "profile": user_profile_2
        })

        query_params_create = {"activate": True}

        # Create Users
        user, resp, err = await client.create_user(
            create_user_req, query_params_create)
        assert err is None

        user_2, resp, err = await client.create_user(
            create_user_req_2, query_params_create)
        assert err is None

        # Create app users
        app_user_credentials_password = models.AppUserPasswordCredential({
            "value": "Password150kta"
        })
        app_user_credentials = models.AppUserCredentials({
            "password": app_user_credentials_password,
            "userName": user.profile.email
        })
        app_user = models.AppUser({
            "credentials": app_user_credentials,
            "id": user.id
        })

        app_user_credentials_password = models.AppUserPasswordCredential({
            "value": "Password150kta"
        })
        app_user_credentials_2 = models.AppUserCredentials({
            "password": app_user_credentials_password,
            "userName": user_2.profile.email
        })
        app_user_2 = models.AppUser({
            "credentials": app_user_credentials_2,
            "id": user_2.id
        })

        # Assign
        created_app_user, _, err = await\
            client.assign_user_to_application(app.id, app_user)
        assert err is None
        assert isinstance(created_app_user, models.AppUser)

        created_app_user_2, _, err = await\
            client.assign_user_to_application(app.id, app_user_2)
        assert err is None
        assert isinstance(created_app_user_2, models.AppUser)

        # Get one user and verify details
        found_app_user, _, err = await\
            client.get_application_user(app.id, user.id)
        assert err is None
        assert found_app_user.id == user.id
        assert found_app_user.scope == "USER"

        # List users in app and verify details
        app_user_list, _, err = await client.list_application_users(app.id)
        assert err is None
        assert len(app_user_list) == 2
        assert next((app_usr for app_usr in app_user_list
                     if app_usr.credentials.user_name == user.profile.email))
        assert next((app_usr for app_usr in app_user_list
                     if app_usr.credentials.user_name == user_2.profile.email))

        # Deactivate and Delete Users
        _, err = await client.deactivate_user(user.id)
        assert err is None
        _, err = await client.deactivate_or_delete_user(user.id)
        assert err is None
        _, err = await client.deactivate_user(user_2.id)
        assert err is None
        _, err = await client.deactivate_or_delete_user(user_2.id)
        assert err is None

        # Deactivate & Delete created app
        _, err = await client.deactivate_application(app.id)
        assert err is None
        _, err = await client.delete_application(app.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_app_update_assigned_user(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Bookmark Application Object
        APP_URL = "https://example.com/bookmark.htm"
        APP_LABEL = "AddBookmarkApp"
        app_settings_app = models.BookmarkApplicationSettingsApplication({
            "requestIntegration": False,
            "url": APP_URL
        })
        app_settings = models.BookmarkApplicationSettings({
            "app": app_settings_app
        })
        bookmark_app_obj = models.BookmarkApplication({
            "label": APP_LABEL,
            "settings": app_settings
        })

        # Create App in org
        app, _, err = await client.create_application(bookmark_app_obj)
        assert err is None
        assert isinstance(app, models.Application)
        assert isinstance(app, models.BookmarkApplication)

        # Create Password
        password = models.PasswordCredential({
            "value": "Password150kta"
        })

        # Create User Credentials
        user_creds = models.UserCredentials({
            "password": password
        })

        # Create User Profile and CreateUser Request
        user_profile = models.UserProfile()
        user_profile.first_name = "John"
        user_profile.last_name = "Doe-Get"
        user_profile.email = "John.Doe-Get@example.com"
        user_profile.login = "John.Doe-Get@example.com"

        create_user_req = models.CreateUserRequest({
            "credentials": user_creds,
            "profile": user_profile
        })

        query_params_create = {"activate": True}

        # Create User
        user, resp, err = await client.create_user(
            create_user_req, query_params_create)
        assert err is None

        # Create app user
        app_user_credentials_password = models.AppUserPasswordCredential({
            "value": "Password150kta"
        })
        app_user_credentials = models.AppUserCredentials({
            "password": app_user_credentials_password,
            "userName": user.profile.email
        })
        app_user = models.AppUser({
            "credentials": app_user_credentials,
            "id": user.id
        })

        # Assign
        created_app_user, _, err = await\
            client.assign_user_to_application(app.id, app_user)
        assert err is None
        assert isinstance(created_app_user, models.AppUser)
        assert created_app_user.scope == "USER"
        assert created_app_user.status == "ACTIVE"
        assert created_app_user.sync_state == "DISABLED"

        # Retrieve
        found_app_user, _, err = await\
            client.get_application_user(app.id, user.id)
        assert err is None
        assert found_app_user.id == user.id

        # Update
        UPDATED_USER_NAME = "JohnJohnJohn"
        UPDATED_PASSWORD = "Password12345!"
        found_app_user.credentials.user_name = UPDATED_USER_NAME
        found_app_user.credentials.password = models.AppUserPasswordCredential(
            {"value": UPDATED_PASSWORD})
        updated_app_user, _, err = await \
            client.update_application_user(app.id, user.id, found_app_user)
        assert updated_app_user.credentials.user_name == UPDATED_USER_NAME

        # Deactivate and Delete User
        _, err = await client.deactivate_user(user.id)
        assert err is None
        _, err = await client.deactivate_or_delete_user(user.id)
        assert err is None

        # Deactivate & Delete created app
        _, err = await client.deactivate_application(app.id)
        assert err is None
        _, err = await client.delete_application(app.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_app_remove_assigned_user(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Bookmark Application Object
        APP_URL = "https://example.com/bookmark.htm"
        APP_LABEL = "AddBookmarkApp"
        app_settings_app = models.BookmarkApplicationSettingsApplication({
            "requestIntegration": False,
            "url": APP_URL
        })
        app_settings = models.BookmarkApplicationSettings({
            "app": app_settings_app
        })
        bookmark_app_obj = models.BookmarkApplication({
            "label": APP_LABEL,
            "settings": app_settings
        })

        # Create App in org
        app, _, err = await client.create_application(bookmark_app_obj)
        assert err is None
        assert isinstance(app, models.Application)
        assert isinstance(app, models.BookmarkApplication)

        # Create Password
        password = models.PasswordCredential({
            "value": "Password150kta"
        })

        # Create User Credentials
        user_creds = models.UserCredentials({
            "password": password
        })

        # Create User Profile and CreateUser Request
        user_profile = models.UserProfile()
        user_profile.first_name = "John"
        user_profile.last_name = "Doe-Get"
        user_profile.email = "John.Doe-Get@example.com"
        user_profile.login = "John.Doe-Get@example.com"

        create_user_req = models.CreateUserRequest({
            "credentials": user_creds,
            "profile": user_profile
        })

        query_params_create = {"activate": True}

        # Create User
        user, resp, err = await client.create_user(
            create_user_req, query_params_create)
        assert err is None

        # Create app user
        app_user_credentials_password = models.AppUserPasswordCredential({
            "value": "Password150kta"
        })
        app_user_credentials = models.AppUserCredentials({
            "password": app_user_credentials_password,
            "userName": user.profile.email
        })
        app_user = models.AppUser({
            "credentials": app_user_credentials,
            "id": user.id
        })

        # Assign
        created_app_user, _, err = await\
            client.assign_user_to_application(app.id, app_user)
        assert err is None
        assert isinstance(created_app_user, models.AppUser)
        assert created_app_user.scope == "USER"
        assert created_app_user.status == "ACTIVE"
        assert created_app_user.sync_state == "DISABLED"

        # Retrieve
        found_app_user, _, err = await\
            client.get_application_user(app.id, user.id)
        assert err is None
        assert found_app_user.id == user.id

        # Remove
        _, err = await client.delete_application_user(app.id, user.id)
        assert err is None

        # Ensure user is not in list
        users_assigned, _, err = await client.list_application_users(app.id)
        assert err is None
        assert len(users_assigned) == 0

        # Deactivate and Delete User
        _, err = await client.deactivate_user(user.id)
        assert err is None
        _, err = await client.deactivate_or_delete_user(user.id)
        assert err is None

        # Deactivate & Delete created app
        _, err = await client.deactivate_application(app.id)
        assert err is None
        _, err = await client.delete_application(app.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_assign_group_app(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Bookmark Application Object
        APP_URL = "https://example.com/bookmark.htm"
        APP_LABEL = "AddBookmarkApp"
        app_settings_app = models.BookmarkApplicationSettingsApplication({
            "requestIntegration": False,
            "url": APP_URL
        })
        app_settings = models.BookmarkApplicationSettings({
            "app": app_settings_app
        })
        bookmark_app_obj = models.BookmarkApplication({
            "label": APP_LABEL,
            "settings": app_settings
        })

        # Create App in org
        app, _, err = await client.create_application(bookmark_app_obj)
        assert err is None
        assert isinstance(app, models.Application)
        assert isinstance(app, models.BookmarkApplication)

        # Create Group Object
        GROUP_NAME = "Group-Target-Test"
        group_profile = models.GroupProfile({
            "name": GROUP_NAME
        })
        group_obj = models.Group({
            "profile": group_profile
        })

        # Create Group
        group, _, err = await client.create_group(group_obj)
        assert err is None
        assert isinstance(group, models.Group)

        # Assign app and group
        assign_ag_req = models.ApplicationGroupAssignment({
            "priority": 0,
            "applicationId": app.id,
            "groupId": group.id
        })

        assign_app_group, _, err = await \
            client.create_application_group_assignment(
                app.id, group.id, assign_ag_req)
        assert err is None

        # 3 second sleep for backend to update

        found_app_group, _, err = await \
            client.get_application_group_assignment(app.id, group.id)
        assert err is None
        assert found_app_group.id == assign_app_group.id
        assert found_app_group.priority == 0

        # Remove and deactivate+delete
        _, err = await client.delete_group(group.id)
        assert err is None

        _, err = await client.deactivate_application(app.id)
        assert err is None
        _, err = await client.delete_application(app.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_list_assign_group_app(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Bookmark Application Object
        APP_URL = "https://example.com/bookmark.htm"
        APP_LABEL = "AddBookmarkApp"
        app_settings_app = models.BookmarkApplicationSettingsApplication({
            "requestIntegration": False,
            "url": APP_URL
        })
        app_settings = models.BookmarkApplicationSettings({
            "app": app_settings_app
        })
        bookmark_app_obj = models.BookmarkApplication({
            "label": APP_LABEL,
            "settings": app_settings
        })

        # Create App in org
        app, _, err = await client.create_application(bookmark_app_obj)
        assert err is None
        assert isinstance(app, models.Application)
        assert isinstance(app, models.BookmarkApplication)

        # Create Group Object
        GROUP_NAME = "Group-Target-Test"
        group_profile = models.GroupProfile({
            "name": GROUP_NAME
        })
        group_obj = models.Group({
            "profile": group_profile
        })

        GROUP_NAME_2 = "Group-Target-Test-2"
        group_profile_2 = models.GroupProfile({
            "name": GROUP_NAME_2
        })
        group_obj_2 = models.Group({
            "profile": group_profile_2
        })

        # Create Groups
        group, _, err = await client.create_group(group_obj)
        assert err is None
        assert isinstance(group, models.Group)
        group_2, _, err = await client.create_group(group_obj_2)
        assert err is None
        assert isinstance(group, models.Group)

        # Assign app and group
        assign_ag_req = models.ApplicationGroupAssignment({
            "priority": 0,
            "applicationId": app.id,
            "groupId": group.id
        })
        assign_ag_req_2 = models.ApplicationGroupAssignment({
            "priority": 0,
            "applicationId": app.id,
            "groupId": group_2.id
        })

        assign_app_group, _, err = await \
            client.create_application_group_assignment(
                app.id, group.id, assign_ag_req)
        assert err is None
        assign_app_group_2, _, err = await \
            client.create_application_group_assignment(
                app.id, group_2.id, assign_ag_req_2)
        assert err is None

        # 3 second sleep for backend to update

        group_assign_list, _, err = await \
            client.list_application_group_assignments(app.id)
        assert err is None
        assert len(group_assign_list) == 2
        assert next(
            (grp for grp in group_assign_list
             if grp.id == assign_app_group.id))
        assert next(
            (grp for grp in group_assign_list
             if grp.id == assign_app_group_2.id))

        # Remove and deactivate+delete
        _, err = await client.delete_group(group.id)
        assert err is None
        _, err = await client.delete_group(group_2.id)
        assert err is None

        _, err = await client.deactivate_application(app.id)
        assert err is None
        _, err = await client.delete_application(app.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_remove_assign_group_app(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Bookmark Application Object
        APP_URL = "https://example.com/bookmark.htm"
        APP_LABEL = "AddBookmarkApp"
        app_settings_app = models.BookmarkApplicationSettingsApplication({
            "requestIntegration": False,
            "url": APP_URL
        })
        app_settings = models.BookmarkApplicationSettings({
            "app": app_settings_app
        })
        bookmark_app_obj = models.BookmarkApplication({
            "label": APP_LABEL,
            "settings": app_settings
        })

        # Create App in org
        app, _, err = await client.create_application(bookmark_app_obj)
        assert err is None
        assert isinstance(app, models.Application)
        assert isinstance(app, models.BookmarkApplication)

        # Create Group Object
        GROUP_NAME = "Group-Target-Test"
        group_profile = models.GroupProfile({
            "name": GROUP_NAME
        })
        group_obj = models.Group({
            "profile": group_profile
        })

        # Create Group
        group, _, err = await client.create_group(group_obj)
        assert err is None
        assert isinstance(group, models.Group)

        # Assign app and group
        assign_ag_req = models.ApplicationGroupAssignment({
            "priority": 0,
            "applicationId": app.id,
            "groupId": group.id
        })

        assign_app_group, _, err = await \
            client.create_application_group_assignment(
                app.id, group.id, assign_ag_req)
        assert err is None

        # 3 second sleep for backend to update

        found_app_group, _, err = await \
            client.get_application_group_assignment(app.id, group.id)
        assert err is None
        assert found_app_group.id == assign_app_group.id
        assert found_app_group.priority == 0

        # Remove
        _, err = await \
            client.delete_application_group_assignment(app.id, group.id)
        assert err is None

        # Verify gone
        group_assign_list, _, err = await \
            client.list_application_group_assignments(app.id)
        assert err is None
        assert len(group_assign_list) == 0

        # Remove and deactivate+delete
        _, err = await client.delete_group(group.id)
        assert err is None

        _, err = await client.deactivate_application(app.id)
        assert err is None
        _, err = await client.delete_application(app.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_list_app_keys(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Bookmark Application Object
        APP_URL = "https://example.com/bookmark.htm"
        APP_LABEL = "AddBookmarkApp"
        app_settings_app = models.BookmarkApplicationSettingsApplication({
            "requestIntegration": False,
            "url": APP_URL
        })
        app_settings = models.BookmarkApplicationSettings({
            "app": app_settings_app
        })
        bookmark_app_obj = models.BookmarkApplication({
            "label": APP_LABEL,
            "settings": app_settings
        })

        # Create App in org
        app, _, err = await client.create_application(bookmark_app_obj)
        assert err is None
        assert isinstance(app, models.Application)
        assert isinstance(app, models.BookmarkApplication)

        # List app keys
        app_keys, _, err = await client.list_application_keys(app.id)
        assert err is None
        assert len(app_keys) == 1

        _, err = await client.deactivate_application(app.id)
        assert err is None
        _, err = await client.delete_application(app.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_get_app_keys(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Bookmark Application Object
        APP_URL = "https://example.com/bookmark.htm"
        APP_LABEL = "AddBookmarkApp"
        app_settings_app = models.BookmarkApplicationSettingsApplication({
            "requestIntegration": False,
            "url": APP_URL
        })
        app_settings = models.BookmarkApplicationSettings({
            "app": app_settings_app
        })
        bookmark_app_obj = models.BookmarkApplication({
            "label": APP_LABEL,
            "settings": app_settings
        })

        # Create App in org
        app, _, err = await client.create_application(bookmark_app_obj)
        assert err is None
        assert isinstance(app, models.Application)
        assert isinstance(app, models.BookmarkApplication)

        # List app keys
        app_keys, _, err = await client.list_application_keys(app.id)
        assert err is None
        assert len(app_keys) == 1

        # Get default key
        default_key = app_keys[0]

        # Find key and verify details
        found_key, _, err = await \
            client.get_application_key(app.id, default_key.kid)
        assert err is None
        assert found_key.kid == default_key.kid
        assert found_key.created == default_key.created
        assert found_key.expires_at == default_key.expires_at
        assert found_key.x_5_c == default_key.x_5_c

        _, err = await client.deactivate_application(app.id)
        assert err is None
        _, err = await client.delete_application(app.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_generate_app_key(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Bookmark Application Object
        APP_URL = "https://example.com/bookmark.htm"
        APP_LABEL = "AddBookmarkApp"
        app_settings_app = models.BookmarkApplicationSettingsApplication({
            "requestIntegration": False,
            "url": APP_URL
        })
        app_settings = models.BookmarkApplicationSettings({
            "app": app_settings_app
        })
        bookmark_app_obj = models.BookmarkApplication({
            "label": APP_LABEL,
            "settings": app_settings
        })

        # Create App in org
        app, _, err = await client.create_application(bookmark_app_obj)
        assert err is None
        assert isinstance(app, models.Application)
        assert isinstance(app, models.BookmarkApplication)

        # Generate key
        query_params_generate = {"validityYears": 2}
        generated, _, err = await\
            client.generate_application_key(app.id, query_params_generate)
        assert err is None

        # List app keys
        found, _, err = await\
            client.get_application_key(app.id, generated.kid)
        assert err is None
        assert found.created == generated.created
        assert found.expires_at == generated.expires_at
        assert found.x_5_c == generated.x_5_c

        _, err = await client.deactivate_application(app.id)
        assert err is None
        _, err = await client.delete_application(app.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_clone_app_key(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Bookmark Application Object
        APP_URL = "https://example.com/bookmark.htm"
        APP_LABEL = "AddBookmarkApp"
        app_settings_app = models.BookmarkApplicationSettingsApplication({
            "requestIntegration": False,
            "url": APP_URL
        })
        app_settings = models.BookmarkApplicationSettings({
            "app": app_settings_app
        })
        bookmark_app_obj = models.BookmarkApplication({
            "label": APP_LABEL,
            "settings": app_settings
        })

        APP_URL_2 = "https://example.com/bookmark2.htm"
        APP_LABEL_2 = "AddBookmarkApp2"
        app_settings_app_2 = models.BookmarkApplicationSettingsApplication({
            "requestIntegration": False,
            "url": APP_URL_2
        })
        app_settings_2 = models.BookmarkApplicationSettings({
            "app": app_settings_app_2
        })
        bookmark_app_obj_2 = models.BookmarkApplication({
            "label": APP_LABEL_2,
            "settings": app_settings_2
        })

        # Create App in org
        app, _, err = await client.create_application(bookmark_app_obj)
        assert err is None
        assert isinstance(app, models.Application)
        assert isinstance(app, models.BookmarkApplication)

        app_2, _, err = await client.create_application(bookmark_app_obj_2)
        assert err is None
        assert isinstance(app, models.Application)
        assert isinstance(app, models.BookmarkApplication)

        # Generate key
        query_params_generate = {"validityYears": 2}
        generated, _, err = await\
            client.generate_application_key(app.id, query_params_generate)
        assert err is None

        # Clone key into second app
        query_params_clone = {"targetAid": app_2.id}
        cloned, _, err = await\
            client.clone_application_key(
                app.id, generated.kid, query_params_clone)
        assert err is None
        assert cloned.kid == generated.kid
        assert cloned.expires_at == generated.expires_at
        assert cloned.x_5_c == generated.x_5_c

        _, err = await client.deactivate_application(app.id)
        assert err is None
        _, err = await client.delete_application(app.id)
        assert err is None
        _, err = await client.deactivate_application(app_2.id)
        assert err is None
        _, err = await client.delete_application(app_2.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_generate_csr(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create CSR
        csr_metadata = models.CsrMetadata({
            "subject": models.CsrMetadataSubject({
                "commonName": "SP Issuer",
                "countryName": "Canada",
                "localityName": "Toronto",
                "organizationName": "Okta, Inc.",
                "organizationalUnitName": "Dev"
            }),
            "subjectAltNames": models.CsrMetadataSubjectAltNames({
                "dnsNames": ["dev.okta.com"]
            })
        })

        # Create Bookmark Application Object
        APP_URL = "https://example.com/bookmark.htm"
        APP_LABEL = "AddBookmarkApp"
        app_settings_app = models.BookmarkApplicationSettingsApplication({
            "requestIntegration": False,
            "url": APP_URL
        })
        app_settings = models.BookmarkApplicationSettings({
            "app": app_settings_app
        })
        bookmark_app_obj = models.BookmarkApplication({
            "label": APP_LABEL,
            "settings": app_settings
        })

        # Create App in org
        app, _, err = await client.create_application(bookmark_app_obj)
        assert err is None
        assert isinstance(app, models.Application)
        assert isinstance(app, models.BookmarkApplication)

        # Generate CSR
        generated, _, err = await \
            client.generate_csr_for_application(app.id, csr_metadata)
        assert err is None
        assert generated.kty == "RSA"
        assert generated.csr is not None

        # Deactivate + delete app
        _, err = await client.deactivate_application(app.id)
        assert err is None
        _, err = await client.delete_application(app.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_get_csr(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create CSR
        csr_metadata = models.CsrMetadata({
            "subject": models.CsrMetadataSubject({
                "commonName": "SP Issuer",
                "countryName": "Canada",
                "localityName": "Toronto",
                "organizationName": "Okta, Inc.",
                "organizationalUnitName": "Dev"
            }),
            "subjectAltNames": models.CsrMetadataSubjectAltNames({
                "dnsNames": ["dev.okta.com"]
            })
        })

        # Create Bookmark Application Object
        APP_URL = "https://example.com/bookmark.htm"
        APP_LABEL = "AddBookmarkApp"
        app_settings_app = models.BookmarkApplicationSettingsApplication({
            "requestIntegration": False,
            "url": APP_URL
        })
        app_settings = models.BookmarkApplicationSettings({
            "app": app_settings_app
        })
        bookmark_app_obj = models.BookmarkApplication({
            "label": APP_LABEL,
            "settings": app_settings
        })

        # Create App in org
        app, _, err = await client.create_application(bookmark_app_obj)
        assert err is None
        assert isinstance(app, models.Application)
        assert isinstance(app, models.BookmarkApplication)

        # Generate CSR
        generated, _, err = await \
            client.generate_csr_for_application(app.id, csr_metadata)
        assert err is None
        assert generated.kty == "RSA"
        assert generated.csr is not None

        # Get CSR
        found, _, err = await client.get_csr_for_application(app.id,
                                                             generated.id)
        assert err is None
        assert found is not None

        # Deactivate + delete app
        _, err = await client.deactivate_application(app.id)
        assert err is None
        _, err = await client.delete_application(app.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_revoke_csr(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create CSR
        csr_metadata = models.CsrMetadata({
            "subject": models.CsrMetadataSubject({
                "commonName": "SP Issuer",
                "countryName": "Canada",
                "localityName": "Toronto",
                "organizationName": "Okta, Inc.",
                "organizationalUnitName": "Dev"
            }),
            "subjectAltNames": models.CsrMetadataSubjectAltNames({
                "dnsNames": ["dev.okta.com"]
            })
        })

        # Create Bookmark Application Object
        APP_URL = "https://example.com/bookmark.htm"
        APP_LABEL = "AddBookmarkApp"
        app_settings_app = models.BookmarkApplicationSettingsApplication({
            "requestIntegration": False,
            "url": APP_URL
        })
        app_settings = models.BookmarkApplicationSettings({
            "app": app_settings_app
        })
        bookmark_app_obj = models.BookmarkApplication({
            "label": APP_LABEL,
            "settings": app_settings
        })

        # Create App in org
        app, _, err = await client.create_application(bookmark_app_obj)
        assert err is None
        assert isinstance(app, models.Application)
        assert isinstance(app, models.BookmarkApplication)

        # Generate CSR
        generated, _, err = await \
            client.generate_csr_for_application(app.id, csr_metadata)
        assert err is None
        assert generated.kty == "RSA"
        assert generated.csr is not None

        # Get CSR
        found, _, err = await client.get_csr_for_application(app.id,
                                                             generated.id)
        assert err is None
        assert found is not None

        # Revoke
        _, err = await client.revoke_csr_from_application(app.id, generated.id)
        assert err is None

        # Verify deletion
        csr_list, _, err = await client.list_csrs_for_application(app.id)
        assert err is None
        assert next((csr for csr in csr_list if csr.id == generated.id), None)\
            is None

        # Deactivate + delete app
        _, err = await client.deactivate_application(app.id)
        assert err is None
        _, err = await client.delete_application(app.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_grant_consent_to_scope(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create OIDC Application Object
        APP_LABEL = "AddOpenIdConnectApp-Grant-Consent-SCOPE"
        CLIENT_URI = "https://example.com/client"
        LOGO_URI = "https://example.com/assets/images/logo-new.png"
        POLICY_URI = "https://example.com/client/policy"
        GRANT_TYPES = [models.OAuthGrantType.AUTHORIZATION_CODE,
                       models.OAuthGrantType.IMPLICIT]
        APP_TYPE = models.OpenIdConnectApplicationType.NATIVE
        POST_LOGOUT_REDIRECT_URIS = ["https://example.com/postlogout",
                                     "myapp://postlogoutcallback"]
        REDIRECT_URIS = ["https://example.com/oauth2/callback",
                         "myapp://callback"]
        RESPONSE_TYPES = [models.OAuthResponseType.TOKEN,
                          models.OAuthResponseType.ID_TOKEN,
                          models.OAuthResponseType.CODE]
        TOS_URL = "https://example.com/client/tos"
        AUTO_KEY_ROTATION = True
        CLIENT_ID = "testclientid12345"
        TOKEN_ENDPOINT_AUTH_METHOD =\
            models.OAuthEndpointAuthenticationMethod.CLIENT_SECRET_POST

        app_settings_client = models.OpenIdConnectApplicationSettingsClient({
            "applicationType": APP_TYPE,
            "clientUri": CLIENT_URI,
            "grantTypes": GRANT_TYPES,
            "logoUri": LOGO_URI,
            "policyUri": POLICY_URI,
            "postLogoutRedirectUris": POST_LOGOUT_REDIRECT_URIS,
            "redirectUris": REDIRECT_URIS,
            "responseTypes": RESPONSE_TYPES,
            "tosUri": TOS_URL
        })
        app_settings = models.OpenIdConnectApplicationSettings({
            "oauthClient": app_settings_client
        })
        app_credentials = models.OAuthApplicationCredentials({
            "oauthClient": models.ApplicationCredentialsOAuthClient({
                "autoKeyRotation": AUTO_KEY_ROTATION,
                "clientId": CLIENT_ID,
                "tokenEndpointAuthMethod": TOKEN_ENDPOINT_AUTH_METHOD
            })
        })
        oidc_app_obj = models.OpenIdConnectApplication({
            "label": APP_LABEL,
            "settings": app_settings,
            "credentials": app_credentials
        })

        # Create App in org
        app, _, err = await client.create_application(oidc_app_obj)
        assert err is None
        assert isinstance(app, models.Application)
        assert isinstance(app, models.OpenIdConnectApplication)

        # Create grant request
        issuer = client.get_base_url()
        SCOPE_ID = "okta.roles.read"
        oauth_scope_consent_grant = models.OAuth2ScopeConsentGrant({
            "issuer": issuer,
            "scopeId": SCOPE_ID
        })

        # Grant consent
        _, _, err = await client.grant_consent_to_scope(
            app.id,
            oauth_scope_consent_grant)
        assert err is None

        # Check grant
        grants, _, err = await client.list_scope_consent_grants(app.id)
        assert err is None
        assert grants is not None
        assert next((grant for grant in grants if grant.scope_id == SCOPE_ID
                     and grant.issuer == issuer))

        # Deactivate & Delete created app
        _, err = await client.deactivate_application(app.id)
        assert err is None
        _, err = await client.delete_application(app.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_get_consent_grant(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create OIDC Application Object
        APP_LABEL = "AddOpenIdConnectApp-Get-Consent-Grant"
        CLIENT_URI = "https://example.com/client"
        LOGO_URI = "https://example.com/assets/images/logo-new.png"
        POLICY_URI = "https://example.com/client/policy"
        GRANT_TYPES = [models.OAuthGrantType.AUTHORIZATION_CODE,
                       models.OAuthGrantType.IMPLICIT]
        APP_TYPE = models.OpenIdConnectApplicationType.NATIVE
        POST_LOGOUT_REDIRECT_URIS = ["https://example.com/postlogout",
                                     "myapp://postlogoutcallback"]
        REDIRECT_URIS = ["https://example.com/oauth2/callback",
                         "myapp://callback"]
        RESPONSE_TYPES = [models.OAuthResponseType.TOKEN,
                          models.OAuthResponseType.ID_TOKEN,
                          models.OAuthResponseType.CODE]
        TOS_URL = "https://example.com/client/tos"
        AUTO_KEY_ROTATION = True
        CLIENT_ID = "testclientid123456"
        TOKEN_ENDPOINT_AUTH_METHOD =\
            models.OAuthEndpointAuthenticationMethod.CLIENT_SECRET_POST

        app_settings_client = models.OpenIdConnectApplicationSettingsClient({
            "applicationType": APP_TYPE,
            "clientUri": CLIENT_URI,
            "grantTypes": GRANT_TYPES,
            "logoUri": LOGO_URI,
            "policyUri": POLICY_URI,
            "postLogoutRedirectUris": POST_LOGOUT_REDIRECT_URIS,
            "redirectUris": REDIRECT_URIS,
            "responseTypes": RESPONSE_TYPES,
            "tosUri": TOS_URL
        })
        app_settings = models.OpenIdConnectApplicationSettings({
            "oauthClient": app_settings_client
        })
        app_credentials = models.OAuthApplicationCredentials({
            "oauthClient": models.ApplicationCredentialsOAuthClient({
                "autoKeyRotation": AUTO_KEY_ROTATION,
                "clientId": CLIENT_ID,
                "tokenEndpointAuthMethod": TOKEN_ENDPOINT_AUTH_METHOD
            })
        })
        oidc_app_obj = models.OpenIdConnectApplication({
            "label": APP_LABEL,
            "settings": app_settings,
            "credentials": app_credentials
        })

        # Create App in org
        app, _, err = await client.create_application(oidc_app_obj)
        assert err is None
        assert isinstance(app, models.Application)
        assert isinstance(app, models.OpenIdConnectApplication)

        # Create grant request
        issuer = client.get_base_url()
        SCOPE_ID = "okta.roles.read"
        oauth_scope_consent_grant = models.OAuth2ScopeConsentGrant({
            "issuer": issuer,
            "scopeId": SCOPE_ID
        })

        # Grant consent
        _, _, err = await client.grant_consent_to_scope(
            app.id,
            oauth_scope_consent_grant)
        assert err is None

        # Check grant
        grants, _, err = await client.list_scope_consent_grants(app.id)
        assert err is None
        assert grants is not None
        found_grant = next((grant for grant in grants if grant.scope_id
                            == SCOPE_ID and grant.issuer == issuer))
        assert found_grant
        found, _, err = await\
            client.get_scope_consent_grant(app.id, found_grant.id)
        assert err is None
        assert found is not None

        # Deactivate & Delete created app
        _, err = await client.deactivate_application(app.id)
        assert err is None
        _, err = await client.delete_application(app.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_revoke_consent_grant(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create OIDC Application Object
        APP_LABEL = "AddOpenIdConnectApp-Revoke-Consent"
        CLIENT_URI = "https://example.com/client"
        LOGO_URI = "https://example.com/assets/images/logo-new.png"
        POLICY_URI = "https://example.com/client/policy"
        GRANT_TYPES = [models.OAuthGrantType.AUTHORIZATION_CODE,
                       models.OAuthGrantType.IMPLICIT]
        APP_TYPE = models.OpenIdConnectApplicationType.NATIVE
        POST_LOGOUT_REDIRECT_URIS = ["https://example.com/postlogout",
                                     "myapp://postlogoutcallback"]
        REDIRECT_URIS = ["https://example.com/oauth2/callback",
                         "myapp://callback"]
        RESPONSE_TYPES = [models.OAuthResponseType.TOKEN,
                          models.OAuthResponseType.ID_TOKEN,
                          models.OAuthResponseType.CODE]
        TOS_URL = "https://example.com/client/tos"
        AUTO_KEY_ROTATION = True
        CLIENT_ID = "testclientid123456"
        TOKEN_ENDPOINT_AUTH_METHOD =\
            models.OAuthEndpointAuthenticationMethod.CLIENT_SECRET_POST

        app_settings_client = models.OpenIdConnectApplicationSettingsClient({
            "applicationType": APP_TYPE,
            "clientUri": CLIENT_URI,
            "grantTypes": GRANT_TYPES,
            "logoUri": LOGO_URI,
            "policyUri": POLICY_URI,
            "postLogoutRedirectUris": POST_LOGOUT_REDIRECT_URIS,
            "redirectUris": REDIRECT_URIS,
            "responseTypes": RESPONSE_TYPES,
            "tosUri": TOS_URL
        })
        app_settings = models.OpenIdConnectApplicationSettings({
            "oauthClient": app_settings_client
        })
        app_credentials = models.OAuthApplicationCredentials({
            "oauthClient": models.ApplicationCredentialsOAuthClient({
                "autoKeyRotation": AUTO_KEY_ROTATION,
                "clientId": CLIENT_ID,
                "tokenEndpointAuthMethod": TOKEN_ENDPOINT_AUTH_METHOD
            })
        })
        oidc_app_obj = models.OpenIdConnectApplication({
            "label": APP_LABEL,
            "settings": app_settings,
            "credentials": app_credentials
        })

        # Create App in org
        app, _, err = await client.create_application(oidc_app_obj)
        assert err is None
        assert isinstance(app, models.Application)
        assert isinstance(app, models.OpenIdConnectApplication)

        # Create grant request
        issuer = client.get_base_url()
        SCOPE_ID = "okta.users.read"
        oauth_scope_consent_grant = models.OAuth2ScopeConsentGrant({
            "issuer": issuer,
            "scopeId": SCOPE_ID
        })

        # Grant consent
        _, _, err = await client.grant_consent_to_scope(
            app.id,
            oauth_scope_consent_grant)
        assert err is None

        # Check grant
        grants, _, err = await client.list_scope_consent_grants(app.id)
        assert err is None
        assert grants is not None
        found_grant = next((grant for grant in grants if grant.scope_id
                            == SCOPE_ID and grant.issuer == issuer))
        assert found_grant
        found, _, err = await\
            client.get_scope_consent_grant(app.id, found_grant.id)
        assert err is None
        assert found is not None

        # Revoke
        _, err = await client.revoke_scope_consent_grant(app.id,
                                                         found_grant.id)
        assert err is None

        # Check again
        updated_grants, _, err = await client.list_scope_consent_grants(app.id)
        assert err is None
        found_grant = next((grant for grant in updated_grants if grant.scope_id
                            == SCOPE_ID and grant.issuer == issuer), None)
        assert found_grant is None

        # Deactivate & Delete created app
        _, err = await client.deactivate_application(app.id)
        assert err is None
        _, err = await client.delete_application(app.id)
        assert err is None



from okta.http_client import HTTPClient
from okta.request_executor import RequestExecutor
from okta.cache.no_op_cache import NoOpCache
from okta.user_agent import UserAgent
from okta.client import Client
from okta.oauth import OAuth
from okta.errors.http_error import HTTPError
from okta.exceptions import HTTPException, OktaAPIException
import tests.mocks as mocks
import pytest
import aiohttp
import asyncio

REQUEST_TIMEOUT = mocks.REQUEST_TIMEOUT
ORG_URL = mocks.ORG_URL
API_TOKEN = mocks.API_TOKEN
CLIENT_ID = mocks.CLIENT_ID
SCOPES = mocks.SCOPES
PRIVATE_KEY = mocks.PRIVATE_KEY
GET_USERS_CALL = mocks.GET_USERS_CALL


@pytest.mark.vcr()
@ pytest.mark.asyncio
async def test_client_success_call_SSWS(fs):
    ssws_client = mocks.MockOktaClient(fs)

    req, error = await ssws_client.get_request_executor()\
        .create_request("GET",
                        GET_USERS_CALL,
                        {},
                        {})

    req, res_details, resp_body, error = await ssws_client\
        .get_request_executor().fire_request(req)

    assert error is None
    assert "User-Agent" in req["headers"]
    assert "Authorization" in req["headers"]
    assert req["headers"]["Authorization"].startswith("SSWS")
    assert res_details.status == 200
    assert "application/json" in res_details.headers["Content-Type"]
    assert resp_body is not None


@ pytest.mark.vcr()
@ pytest.mark.asyncio
async def test_client_error_call_SSWS(fs):
    ssws_client = Client({
        "orgUrl": ORG_URL,
        "token": API_TOKEN + "wrong token"
    })

    req, error = await ssws_client.get_request_executor()\
        .create_request("GET",
                        GET_USERS_CALL,
                        {},
                        {})

    req, res_details, resp_body, error = await ssws_client\
        .get_request_executor().fire_request(req)

    parsed, error = HTTPClient.check_response_for_error(
        req["url"], res_details, resp_body)

    assert parsed is None
    assert res_details.status == 404
    assert isinstance(error, HTTPError)
    assert error.message.startswith("HTTP 404")


@pytest.mark.vcr()
@pytest.mark.asyncio
async def test_client_success_call_oauth(fs, monkeypatch):
    oauth_client = Client({
        "orgUrl": ORG_URL,
        "authorizationMode": "PrivateKey",
        "clientId": CLIENT_ID,
        "scopes": SCOPES,
        "privateKey": PRIVATE_KEY
    })

    request_executor = oauth_client.get_request_executor()

    monkeypatch.setattr(OAuth, 'get_access_token', mocks.mock_access_token)
    req, err = await request_executor.create_request(
        "GET",
        GET_USERS_CALL,
        {},
        {})

    req, res_details, resp_body, error = await oauth_client\
        .get_request_executor().fire_request(req)

    assert error is None
    assert "User-Agent" in req["headers"]
    assert "Authorization" in req["headers"]
    assert req["headers"]["Authorization"].startswith("Bearer")
    assert res_details.status == 200
    assert "application/json" in res_details.headers["Content-Type"]
    assert resp_body is not None


@pytest.mark.vcr()
@ pytest.mark.asyncio
async def test_client_error_call_oauth(fs, monkeypatch):
    oauth_client = Client({
        "orgUrl": ORG_URL,
        "authorizationMode": "PrivateKey",
        "clientId": CLIENT_ID,
        "scopes": SCOPES,
        "privateKey": PRIVATE_KEY + "Wrong one"
    })

    monkeypatch.setattr(OAuth, 'get_access_token', mocks.mock_access_token)

    req, err = await oauth_client.get_request_executor()\
        .create_request("GET",
                        GET_USERS_CALL,
                        {},
                        {})

    req, res_details, resp_body, error = await oauth_client\
        .get_request_executor().fire_request(req)

    parsed, error = HTTPClient.check_response_for_error(
        req["url"], res_details, resp_body)

    assert parsed is None
    assert res_details.status == 404
    assert isinstance(error, HTTPError)
    assert error.message.startswith("HTTP 404")


@ pytest.mark.asyncio
async def test_client_invalid_url():
    http_client = HTTPClient({
        'requestTimeout': REQUEST_TIMEOUT,
        'headers': {}
    })
    req, res_details, resp_body, error = await http_client.send_request({
        'method': 'GET',
        'url': "",
        'headers': {},
        'data': {}
    })
    assert all(values in [None] for values in [req, res_details, resp_body])
    assert issubclass(type(error), aiohttp.ClientError)
    assert type(error) == aiohttp.InvalidURL


@ pytest.mark.asyncio
async def test_client_invalid_HTTP_method(monkeypatch):
    http_client = HTTPClient({
        'requestTimeout': REQUEST_TIMEOUT,
        'headers': {}
    })

    monkeypatch.setattr(HTTPClient, 'send_request',
                        mocks.mock_invalid_HTTP_response)

    req, res_details, resp_body, error = await http_client.send_request({
        'method': 'INVALID',
        'url': ORG_URL,
        'headers': {},
        'data': {}
    })
    assert all(values in [None] for values in [req, res_details, resp_body])
    assert issubclass(type(error), aiohttp.ClientError)
    assert type(error) == aiohttp.ContentTypeError

    req, res_details, resp_body, error = await http_client.send_request({
        'method': '',
        'url': ORG_URL,
        'headers': {},
        'data': {}
    })
    assert all(values in [None] for values in [req, res_details, resp_body])
    assert issubclass(type(error), aiohttp.ClientError)
    assert type(error) == aiohttp.ContentTypeError


@ pytest.mark.asyncio
async def test_client_user_agent(monkeypatch):
    req_exec = RequestExecutor({
        "client": {
            "orgUrl": "https://test.okta.com",
            "authorizationMode": 'SSWS',
            "token": 'token',
            "userAgent": '',
            "rateLimit": {
                "maxRetries": 2
            }
        }
    }, NoOpCache(), None)

    assert 'User-Agent' in req_exec._http_client._default_headers
    assert req_exec._http_client._default_headers["User-Agent"] == UserAgent(
    ).get_user_agent_string()

    # mock and shoot request
    monkeypatch.setattr(HTTPClient, 'send_request',
                        mocks.mock_GET_HTTP_Client_response_valid)

    req, res_details, resp_body, error = await req_exec._http_client.\
        send_request({
            'method': 'GET',
            'url': ORG_URL + GET_USERS_CALL,
            'headers': {'Authorization': f"SSWS {API_TOKEN}"},
            'data': {}
        })
    assert error is None
    assert 'User-Agent' in req["headers"]
    assert req_exec._http_client._default_headers["User-Agent"] == \
        req["headers"]["User-Agent"]

    # with extra user agent string
    extra = "testing"
    req_exec = RequestExecutor({
        "client": {
            "orgUrl": "https://test.okta.com",
            "authorizationMode": 'SSWS',
            "token": 'token',
            "userAgent": extra,
            "rateLimit": {
                "maxRetries": 2
            }
        }
    }, NoOpCache(), None)
    assert 'User-Agent' in req_exec._http_client._default_headers
    assert req_exec._http_client._default_headers["User-Agent"] == \
        UserAgent(extra).get_user_agent_string()


@ pytest.mark.asyncio
async def test_client_timeout(monkeypatch):
    http_client = HTTPClient({
        'requestTimeout': REQUEST_TIMEOUT,
        'headers': {}
    })

    # mock and shoot request
    monkeypatch.setattr(HTTPClient, 'send_request',
                        mocks.mock_timeout_response)

    req, res_details, resp_body, error = await http_client.send_request({
        'method': 'INVALID',
        'url': ORG_URL,
        'headers': {},
        'data': {}
    })

    assert all(values in [None] for values in [req, res_details, resp_body])
    assert type(error) == asyncio.TimeoutError


@ pytest.mark.asyncio
async def test_client_error_raise_okta_api_exception(monkeypatch, mocker):
    """Verify http_client raises OktaAPIException in case of API error
    and `raise_exception` attribute is True
    """
    mock_res_details = mocker.Mock()
    mock_res_details.status = 500
    mock_resp_body = {'errorCode': 500,
                      'errorLink': 'https://test_okta_error',
                      'errorId': 'test_id'}
    monkeypatch.setattr(HTTPClient, 'raise_exception', True)

    with pytest.raises(OktaAPIException):
        parsed, error = HTTPClient.check_response_for_error(
            'https://test.okta.com', mock_res_details, mock_resp_body)


@ pytest.mark.asyncio
async def test_client_error_raise_http_exception(monkeypatch, mocker):
    """Verify http_client raises HTTPException in case of HTTP error
    and `raise_exception` attribute is True
    """
    mock_res_details = mocker.Mock()
    mock_res_details.status = 500
    monkeypatch.setattr(HTTPClient, 'raise_exception', True)

    with pytest.raises(HTTPException):
        parsed, error = HTTPClient.check_response_for_error(
            'https://test.okta.com', mock_res_details, {})



import pytest
from tests.mocks import MockOktaClient
import okta.models as models
from http import HTTPStatus
from okta.errors.okta_api_error import OktaAPIError


class TestAuthorizationServerResource:
    """
    Integration Tests for the Authorization Server Resource
    """
    SDK_PREFIX = "python_sdk"

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_create_auth_server(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Auth Server
        TEST_NAME = f"{TestAuthorizationServerResource.SDK_PREFIX}_test_aaPeZ"
        TEST_DESC = "Test Auth Server"
        TEST_AUDS = ["api://default"]
        auth_server_model = models.AuthorizationServer({
            "name": TEST_NAME,
            "description": TEST_DESC,
            "audiences": TEST_AUDS
        })

        created_auth_server, _, err = await \
            client.create_authorization_server(auth_server_model)
        assert err is None
        assert isinstance(created_auth_server, models.AuthorizationServer)
        assert created_auth_server.name == TEST_NAME
        assert created_auth_server.description == TEST_DESC
        assert created_auth_server.audiences == TEST_AUDS
        assert created_auth_server.audiences[0] == TEST_AUDS[0]

        # Clean up
        _, err = await client.deactivate_authorization_server(
            created_auth_server.id)
        assert err is None
        _, err = await client.delete_authorization_server(
            created_auth_server.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_list_auth_servers(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Auth Server
        TEST_NAME = f"{TestAuthorizationServerResource.SDK_PREFIX}_test_aaPeZ"
        TEST_DESC = "Test Auth Server"
        TEST_AUDS = ["api://default"]
        auth_server_model = models.AuthorizationServer({
            "name": TEST_NAME,
            "description": TEST_DESC,
            "audiences": TEST_AUDS
        })

        created_auth_server, _, err = await \
            client.create_authorization_server(auth_server_model)
        assert err is None
        assert isinstance(created_auth_server, models.AuthorizationServer)
        assert created_auth_server.name == TEST_NAME
        assert created_auth_server.description == TEST_DESC
        assert created_auth_server.audiences == TEST_AUDS
        assert created_auth_server.audiences[0] == TEST_AUDS[0]

        # List Auth Servers
        auth_servers, _, err = await client.list_authorization_servers()
        assert err is None
        # Assuming default Auth server instantiated
        assert auth_servers is not None
        assert len(auth_servers) > 0
        assert next(
            (server for server in auth_servers
             if server.id == created_auth_server.id))

        # Clean up
        _, err = await client.deactivate_authorization_server(
            created_auth_server.id)
        assert err is None
        _, err = await client.delete_authorization_server(
            created_auth_server.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_get_created_auth_server(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Auth Server
        TEST_NAME = f"{TestAuthorizationServerResource.SDK_PREFIX}_test_aaPeZ"
        TEST_DESC = "Test Auth Server"
        TEST_AUDS = ["api://default"]
        auth_server_model = models.AuthorizationServer({
            "name": TEST_NAME,
            "description": TEST_DESC,
            "audiences": TEST_AUDS
        })

        created_auth_server, _, err = await \
            client.create_authorization_server(auth_server_model)
        assert err is None
        assert isinstance(created_auth_server, models.AuthorizationServer)
        assert created_auth_server.name == TEST_NAME
        assert created_auth_server.description == TEST_DESC
        assert created_auth_server.audiences == TEST_AUDS
        assert created_auth_server.audiences[0] == TEST_AUDS[0]

        # Get Auth Server
        retrieved_auth_server, _, err = await\
            client.get_authorization_server(created_auth_server.id)
        assert err is None
        assert retrieved_auth_server.id ==\
            created_auth_server.id
        assert retrieved_auth_server.name ==\
            created_auth_server.name
        assert retrieved_auth_server.description ==\
            created_auth_server.description

        # Clean up
        _, err = await client.deactivate_authorization_server(
            created_auth_server.id)
        assert err is None
        _, err = await client.delete_authorization_server(
            created_auth_server.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_update_auth_server(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Auth Server
        TEST_NAME = f"{TestAuthorizationServerResource.SDK_PREFIX}_test_abPeZ"
        TEST_DESC = "Test Auth Server"
        TEST_AUDS = ["api://default"]
        auth_server_model = models.AuthorizationServer({
            "name": TEST_NAME,
            "description": TEST_DESC,
            "audiences": TEST_AUDS
        })

        created_auth_server, _, err = await \
            client.create_authorization_server(auth_server_model)
        assert err is None
        assert isinstance(created_auth_server, models.AuthorizationServer)
        assert created_auth_server.name == TEST_NAME
        assert created_auth_server.description == TEST_DESC
        assert created_auth_server.audiences == TEST_AUDS
        assert created_auth_server.audiences[0] ==\
            TEST_AUDS[0]

        # Update
        UPDATED_DESC = f"{TEST_DESC}-Updated"
        created_auth_server.description = UPDATED_DESC
        updated_auth_server, _, err = await client.update_authorization_server(
            created_auth_server.id,
            created_auth_server)
        assert err is None
        assert updated_auth_server.description == UPDATED_DESC
        assert updated_auth_server.id == created_auth_server.id

        # Get Auth Server
        retrieved_server, _, err = await\
            client.get_authorization_server(created_auth_server.id)
        assert err is None
        assert retrieved_server.id == updated_auth_server.id
        assert retrieved_server.name == updated_auth_server.name
        assert retrieved_server.description == updated_auth_server.description

        # Clean up
        _, err = await client.deactivate_authorization_server(
            created_auth_server.id)
        assert err is None
        _, err = await client.delete_authorization_server(
            created_auth_server.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_update_deactivate_delete_server(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Auth Server
        TEST_NAME = f"{TestAuthorizationServerResource.SDK_PREFIX}_test_abCeZ"
        TEST_DESC = "Test Auth Server"
        TEST_AUDS = ["api://default"]
        auth_server_model = models.AuthorizationServer({
            "name": TEST_NAME,
            "description": TEST_DESC,
            "audiences": TEST_AUDS
        })

        created_auth_server, _, err = await \
            client.create_authorization_server(auth_server_model)
        assert err is None
        assert isinstance(created_auth_server, models.AuthorizationServer)
        assert created_auth_server.name == TEST_NAME
        assert created_auth_server.description == TEST_DESC
        assert created_auth_server.audiences == TEST_AUDS
        assert created_auth_server.audiences[0] == TEST_AUDS[0]

        # Get Auth Server
        retrieved_auth_server, _, err = await\
            client.get_authorization_server(created_auth_server.id)
        assert err is None
        assert retrieved_auth_server.id == created_auth_server.id
        assert retrieved_auth_server.status == "ACTIVE"

        # DeActivate Auth server
        _, err = await client.deactivate_authorization_server(
            created_auth_server.id)
        assert err is None

        # Get Auth Server
        retrieved_auth_server, _, err = await\
            client.get_authorization_server(created_auth_server.id)
        assert err is None
        assert retrieved_auth_server.id == created_auth_server.id
        assert retrieved_auth_server.status == "INACTIVE"

        # Delete Auth server
        _, err = await client.delete_authorization_server(
            created_auth_server.id)
        assert err is None

        # Get Auth Server
        retrieved_auth_server, resp, err = await\
            client.get_authorization_server(created_auth_server.id)
        assert err is not None
        assert isinstance(err, OktaAPIError)
        assert resp.get_status() == HTTPStatus.NOT_FOUND

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_activate_server(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Auth Server
        TEST_NAME = f"{TestAuthorizationServerResource.SDK_PREFIX}_test_abCeZ"
        TEST_DESC = "Test Auth Server"
        TEST_AUDS = ["api://default"]
        auth_server_model = models.AuthorizationServer({
            "name": TEST_NAME,
            "description": TEST_DESC,
            "audiences": TEST_AUDS
        })

        created_auth_server, _, err = await \
            client.create_authorization_server(auth_server_model)
        assert err is None
        assert isinstance(created_auth_server, models.AuthorizationServer)
        assert created_auth_server.name == TEST_NAME
        assert created_auth_server.description == TEST_DESC
        assert created_auth_server.audiences == TEST_AUDS
        assert created_auth_server.audiences[0] == TEST_AUDS[0]

        # Get Auth Server
        retrieved_auth_server, _, err = await\
            client.get_authorization_server(created_auth_server.id)
        assert err is None
        assert retrieved_auth_server.id == created_auth_server.id
        assert retrieved_auth_server.status == "ACTIVE"

        # DeActivate Auth server
        _, err = await client.deactivate_authorization_server(
            created_auth_server.id)
        assert err is None

        # Get Auth Server
        retrieved_auth_server, _, err = await\
            client.get_authorization_server(created_auth_server.id)
        assert err is None
        assert retrieved_auth_server.id == created_auth_server.id
        assert retrieved_auth_server.status == "INACTIVE"

        # Activate Auth server
        _, err = await client.activate_authorization_server(
            created_auth_server.id)
        assert err is None

        # Get Auth Server
        retrieved_auth_server, _, err = await\
            client.get_authorization_server(created_auth_server.id)
        assert err is None
        assert retrieved_auth_server.id == created_auth_server.id
        assert retrieved_auth_server.status == "ACTIVE"

        # DeActivate Auth server
        _, err = await client.deactivate_authorization_server(
            created_auth_server.id)
        assert err is None
        # Delete Auth server
        _, err = await client.delete_authorization_server(
            created_auth_server.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_list_auth_server_policies(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Auth Server
        TEST_NAME = f"{TestAuthorizationServerResource.SDK_PREFIX}_test_abDeZ"
        TEST_DESC = "Test Auth Server"
        TEST_AUDS = ["api://default"]
        auth_server_model = models.AuthorizationServer({
            "name": TEST_NAME,
            "description": TEST_DESC,
            "audiences": TEST_AUDS
        })

        created_auth_server, _, err = await \
            client.create_authorization_server(auth_server_model)
        assert err is None
        assert isinstance(created_auth_server, models.AuthorizationServer)
        assert created_auth_server.name == TEST_NAME
        assert created_auth_server.description == TEST_DESC
        assert created_auth_server.audiences == TEST_AUDS
        assert created_auth_server.audiences[0] == TEST_AUDS[0]

        # Create Policy
        POLICY_TYPE = models.PolicyType.OAUTH_AUTHORIZATION_POLICY
        POLICY_STATUS = "ACTIVE"
        POLICY_NAME = "Test Policy"
        POLICY_DESC = "Test Policy"
        POLICY_PRIORITY = 1
        POLICY_CONDITIONS = models.PolicyRuleConditions({
            "clients": models.ClientPolicyCondition({
                "include": ["ALL_CLIENTS"]
            })
        })

        policy_model = models.Policy({
            "type": POLICY_TYPE,
            "status": POLICY_STATUS,
            "name": POLICY_NAME,
            "description": POLICY_DESC,
            "priority": POLICY_PRIORITY,
            "conditions": POLICY_CONDITIONS
        })

        created_policy, _, err = await\
            client.create_authorization_server_policy(
                created_auth_server.id, policy_model
            )
        assert err is None

        # Get Policies
        policies, _, err = await client.list_authorization_server_policies(
            created_auth_server.id
        )
        assert err is None
        assert len(policies) > 0
        assert next((plcy for plcy in policies
                     if plcy.id == created_policy.id))

        # Delete Policy
        _, err = await client.delete_authorization_server_policy(
            created_auth_server.id, created_policy.id
        )
        assert err is None

        # DeActivate Auth server
        _, err = await client.deactivate_authorization_server(
            created_auth_server.id)
        assert err is None
        # Delete Auth server
        _, err = await client.delete_authorization_server(
            created_auth_server.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_get_auth_server_policy(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Auth Server
        TEST_NAME = f"{TestAuthorizationServerResource.SDK_PREFIX}_test_abDfZ"
        TEST_DESC = "Test Auth Server"
        TEST_AUDS = ["api://default"]
        auth_server_model = models.AuthorizationServer({
            "name": TEST_NAME,
            "description": TEST_DESC,
            "audiences": TEST_AUDS
        })

        created_auth_server, _, err = await \
            client.create_authorization_server(auth_server_model)
        assert err is None
        assert isinstance(created_auth_server, models.AuthorizationServer)
        assert created_auth_server.name == TEST_NAME
        assert created_auth_server.description == TEST_DESC
        assert created_auth_server.audiences == TEST_AUDS
        assert created_auth_server.audiences[0] == TEST_AUDS[0]

        # Create Policy
        POLICY_TYPE = models.PolicyType.OAUTH_AUTHORIZATION_POLICY
        POLICY_STATUS = "ACTIVE"
        POLICY_NAME = "Test Policy"
        POLICY_DESC = "Test Policy"
        POLICY_PRIORITY = 1
        POLICY_CONDITIONS = models.PolicyRuleConditions({
            "clients": models.ClientPolicyCondition({
                "include": ["ALL_CLIENTS"]
            })
        })

        policy_model = models.Policy({
            "type": POLICY_TYPE,
            "status": POLICY_STATUS,
            "name": POLICY_NAME,
            "description": POLICY_DESC,
            "priority": POLICY_PRIORITY,
            "conditions": POLICY_CONDITIONS
        })

        created_policy, _, err = await\
            client.create_authorization_server_policy(
                created_auth_server.id, policy_model
            )
        assert err is None

        # Get Policy
        found_policy, _, err = await client.get_authorization_server_policy(
            created_auth_server.id, created_policy.id
        )
        assert err is None
        assert found_policy.id == created_policy.id
        assert found_policy.name == created_policy.name
        assert found_policy.description == created_policy.description

        # Delete Policy
        _, err = await client.delete_authorization_server_policy(
            created_auth_server.id, created_policy.id
        )
        assert err is None

        # DeActivate Auth server
        _, err = await client.deactivate_authorization_server(
            created_auth_server.id)
        assert err is None
        # Delete Auth server
        _, err = await client.delete_authorization_server(
            created_auth_server.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_update_auth_server_policy(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Auth Server
        TEST_NAME = f"{TestAuthorizationServerResource.SDK_PREFIX}_test_abDfZ"
        TEST_DESC = "Test Auth Server"
        TEST_AUDS = ["api://default"]
        auth_server_model = models.AuthorizationServer({
            "name": TEST_NAME,
            "description": TEST_DESC,
            "audiences": TEST_AUDS
        })

        auth_server, _, err = await \
            client.create_authorization_server(auth_server_model)
        assert err is None
        assert isinstance(auth_server, models.AuthorizationServer)
        assert auth_server.name == TEST_NAME
        assert auth_server.description == TEST_DESC
        assert auth_server.audiences == TEST_AUDS
        assert auth_server.audiences[0] == TEST_AUDS[0]

        # Create Policy
        POLICY_TYPE = models.PolicyType.OAUTH_AUTHORIZATION_POLICY
        POLICY_STATUS = "ACTIVE"
        POLICY_NAME = "Test Policy"
        POLICY_DESC = "Test Policy"
        POLICY_PRIORITY = 1
        POLICY_CONDITIONS = models.PolicyRuleConditions({
            "clients": models.ClientPolicyCondition({
                "include": ["ALL_CLIENTS"]
            })
        })

        policy_model = models.Policy({
            "type": POLICY_TYPE,
            "status": POLICY_STATUS,
            "name": POLICY_NAME,
            "description": POLICY_DESC,
            "priority": POLICY_PRIORITY,
            "conditions": POLICY_CONDITIONS
        })

        created_policy, _, err = await\
            client.create_authorization_server_policy(
                auth_server.id, policy_model
            )
        assert err is None

        # Update policy
        UPDATED_NAME = f"{POLICY_NAME}_updated"
        UPDATED_DESC = f"{POLICY_DESC}_updated"
        created_policy.name = UPDATED_NAME
        created_policy.description = UPDATED_DESC

        updated_auth_server_policy, _, err = await\
            client.update_authorization_server_policy(
                auth_server.id, created_policy.id, created_policy
            )
        assert err is None
        assert updated_auth_server_policy.name == UPDATED_NAME
        assert updated_auth_server_policy.description == UPDATED_DESC

        # Delete Policy
        _, err = await client.delete_authorization_server_policy(
            auth_server.id, created_policy.id
        )
        assert err is None

        # DeActivate Auth server
        _, err = await client.deactivate_authorization_server(
            auth_server.id)
        assert err is None
        # Delete Auth server
        _, err = await client.delete_authorization_server(
            auth_server.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_delete_auth_server_policy(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Auth Server
        TEST_NAME = f"{TestAuthorizationServerResource.SDK_PREFIX}_test_abDfZ"
        TEST_DESC = "Test Auth Server"
        TEST_AUDS = ["api://default"]
        auth_server_model = models.AuthorizationServer({
            "name": TEST_NAME,
            "description": TEST_DESC,
            "audiences": TEST_AUDS
        })

        created_auth_server, _, err = await \
            client.create_authorization_server(auth_server_model)
        assert err is None
        assert isinstance(created_auth_server, models.AuthorizationServer)
        assert created_auth_server.name == TEST_NAME
        assert created_auth_server.description == TEST_DESC
        assert created_auth_server.audiences == TEST_AUDS
        assert created_auth_server.audiences[0] == TEST_AUDS[0]

        # Create Policy
        POLICY_TYPE = models.PolicyType.OAUTH_AUTHORIZATION_POLICY
        POLICY_STATUS = "ACTIVE"
        POLICY_NAME = "Test Policy"
        POLICY_DESC = "Test Policy"
        POLICY_PRIORITY = 1
        POLICY_CONDITIONS = models.PolicyRuleConditions({
            "clients": models.ClientPolicyCondition({
                "include": ["ALL_CLIENTS"]
            })
        })

        policy_model = models.Policy({
            "type": POLICY_TYPE,
            "status": POLICY_STATUS,
            "name": POLICY_NAME,
            "description": POLICY_DESC,
            "priority": POLICY_PRIORITY,
            "conditions": POLICY_CONDITIONS
        })

        created_policy, _, err = await\
            client.create_authorization_server_policy(
                created_auth_server.id, policy_model
            )
        assert err is None

        # Get Policy
        found_policy, _, err = await\
            client.get_authorization_server_policy(
                created_auth_server.id, created_policy.id
            )
        assert err is None
        assert found_policy is not None
        assert found_policy.id == created_policy.id
        assert found_policy.name == created_policy.name
        assert found_policy.description == created_policy.description

        # Delete Policy
        _, err = await client.delete_authorization_server_policy(
            created_auth_server.id, created_policy.id
        )
        assert err is None

        # Get Policy
        found_policy, resp, err = await\
            client.get_authorization_server_policy(
                created_auth_server.id, created_policy.id
            )
        assert err is not None
        assert isinstance(err, OktaAPIError)
        assert resp.get_status() == HTTPStatus.NOT_FOUND
        assert found_policy is None

        # DeActivate Auth server
        _, err = await client.deactivate_authorization_server(
            created_auth_server.id)
        assert err is None
        # Delete Auth server
        _, err = await client.delete_authorization_server(
            created_auth_server.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_create_get_oauth2_scope(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Auth Server
        TEST_NAME = f"{TestAuthorizationServerResource.SDK_PREFIX}_test_abDfZ"
        TEST_DESC = "Test Auth Server"
        TEST_AUDS = ["api://default"]
        auth_server_model = models.AuthorizationServer({
            "name": TEST_NAME,
            "description": TEST_DESC,
            "audiences": TEST_AUDS
        })

        auth_server, _, err = await \
            client.create_authorization_server(auth_server_model)
        assert err is None
        assert isinstance(auth_server, models.AuthorizationServer)
        assert auth_server.name == TEST_NAME
        assert auth_server.description == TEST_DESC
        assert auth_server.audiences == TEST_AUDS
        assert auth_server.audiences[0] == TEST_AUDS[0]

        # Create Oauth Scope
        SCOPE_NAME = f"{TestAuthorizationServerResource.SDK_PREFIX}:abDGz"
        scope_obj = models.OAuth2Scope({
            "name": SCOPE_NAME
        })

        oauth_scope, _, err = await client.create_o_auth_2_scope(
            auth_server.id, scope_obj
        )
        assert err is None
        assert isinstance(oauth_scope, models.OAuth2Scope)
        assert oauth_scope.name == scope_obj.name

        # Get Oauth Scope
        found_oauth_scope, _, err = await client.get_o_auth_2_scope(
            auth_server.id, oauth_scope.id
        )
        assert err is None
        assert found_oauth_scope.name == oauth_scope.name

        # Delete Oauth Scope
        _, err = await client.delete_o_auth_2_scope(
            auth_server.id, oauth_scope.id
        )
        assert err is None

        # DeActivate Auth server
        _, err = await client.deactivate_authorization_server(auth_server.id)
        assert err is None
        # Delete Auth server
        _, err = await client.delete_authorization_server(auth_server.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_list_oauth2_scopes(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Auth Server
        TEST_NAME = f"{TestAuthorizationServerResource.SDK_PREFIX}_test_abDgZ"
        TEST_DESC = "Test Auth Server"
        TEST_AUDS = ["api://default"]
        auth_server_model = models.AuthorizationServer({
            "name": TEST_NAME,
            "description": TEST_DESC,
            "audiences": TEST_AUDS
        })

        created_auth_server, _, err = await \
            client.create_authorization_server(auth_server_model)
        assert err is None
        assert isinstance(created_auth_server, models.AuthorizationServer)
        assert created_auth_server.name == TEST_NAME
        assert created_auth_server.description == TEST_DESC
        assert created_auth_server.audiences == TEST_AUDS
        assert created_auth_server.audiences[0] == TEST_AUDS[0]

        # Create Oauth Scope
        SCOPE_NAME = f"{TestAuthorizationServerResource.SDK_PREFIX}:abDGz"
        scope_obj = models.OAuth2Scope({
            "name": SCOPE_NAME
        })

        oauth_scope, _, err = await client.create_o_auth_2_scope(
            created_auth_server.id, scope_obj
        )
        assert err is None
        assert isinstance(oauth_scope, models.OAuth2Scope)
        assert oauth_scope.name == scope_obj.name

        # List Oauth Scopes
        scopes, _, err = await client.list_o_auth_2_scopes(
            created_auth_server.id)
        assert err is None
        assert len(scopes) > 0
        assert next((scope for scope in scopes if scope.id == oauth_scope.id))

        # Delete Oauth Scope
        _, err = await client.delete_o_auth_2_scope(
            created_auth_server.id, oauth_scope.id
        )
        assert err is None

        # DeActivate Auth server
        _, err = await client.deactivate_authorization_server(
            created_auth_server.id)
        assert err is None
        # Delete Auth server
        _, err = await client.delete_authorization_server(
            created_auth_server.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_update_oauth2_scope(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Auth Server
        TEST_NAME = f"{TestAuthorizationServerResource.SDK_PREFIX}_test_abDfZ"
        TEST_DESC = "Test Auth Server"
        TEST_AUDS = ["api://default"]
        auth_server_model = models.AuthorizationServer({
            "name": TEST_NAME,
            "description": TEST_DESC,
            "audiences": TEST_AUDS
        })

        created_auth_server, _, err = await \
            client.create_authorization_server(auth_server_model)
        assert err is None
        assert isinstance(created_auth_server, models.AuthorizationServer)
        assert created_auth_server.name == TEST_NAME
        assert created_auth_server.description == TEST_DESC
        assert created_auth_server.audiences == TEST_AUDS
        assert created_auth_server.audiences[0] == TEST_AUDS[0]

        # Create Oauth Scope
        SCOPE_NAME = f"{TestAuthorizationServerResource.SDK_PREFIX}:abDHz"
        scope_obj = models.OAuth2Scope({
            "name": SCOPE_NAME,
            "consent": "REQUIRED",
            "metadataPublish": "ALL_CLIENTS"
        })
        UPDATED_SCOPE_NAME = f"{SCOPE_NAME}updated"
        updated_obj = models.OAuth2Scope({
            "name": UPDATED_SCOPE_NAME,
            "consent": "REQUIRED",
            "metadataPublish": "ALL_CLIENTS"
        })

        oauth_scope, _, err = await client.create_o_auth_2_scope(
            created_auth_server.id, scope_obj
        )
        assert err is None
        assert isinstance(oauth_scope, models.OAuth2Scope)
        assert oauth_scope.name == scope_obj.name

        # Update scope
        updated_scope, _, err = await client.update_o_auth_2_scope(
            created_auth_server.id, oauth_scope.id, updated_obj
        )
        assert err is None
        assert updated_scope.name == UPDATED_SCOPE_NAME

        # Delete Oauth Scope
        _, err = await client.delete_o_auth_2_scope(
            created_auth_server.id, oauth_scope.id
        )
        assert err is None

        # DeActivate Auth server
        _, err = await client.deactivate_authorization_server(
            created_auth_server.id)
        assert err is None
        # Delete Auth server
        _, err = await client.delete_authorization_server(
            created_auth_server.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_delete_oauth2_scope(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Auth Server
        TEST_NAME = f"{TestAuthorizationServerResource.SDK_PREFIX}_test_abDfZ"
        TEST_DESC = "Test Auth Server"
        TEST_AUDS = ["api://default"]
        auth_server_model = models.AuthorizationServer({
            "name": TEST_NAME,
            "description": TEST_DESC,
            "audiences": TEST_AUDS
        })

        created_auth_server, _, err = await \
            client.create_authorization_server(auth_server_model)
        assert err is None
        assert isinstance(created_auth_server, models.AuthorizationServer)
        assert created_auth_server.name == TEST_NAME
        assert created_auth_server.description == TEST_DESC
        assert created_auth_server.audiences == TEST_AUDS
        assert created_auth_server.audiences[0] == TEST_AUDS[0]

        # Create Oauth Scope
        SCOPE_NAME = f"{TestAuthorizationServerResource.SDK_PREFIX}:abDGz"
        scope_obj = models.OAuth2Scope({
            "name": SCOPE_NAME
        })

        oauth_scope, _, err = await client.create_o_auth_2_scope(
            created_auth_server.id, scope_obj
        )
        assert err is None
        assert isinstance(oauth_scope, models.OAuth2Scope)
        assert oauth_scope.name == scope_obj.name

        # Get Oauth Scope
        found_oauth_scope, _, err = await client.get_o_auth_2_scope(
            created_auth_server.id, oauth_scope.id
        )
        assert err is None
        assert found_oauth_scope.name == oauth_scope.name

        # Delete Oauth Scope
        _, err = await client.delete_o_auth_2_scope(
            created_auth_server.id, oauth_scope.id
        )
        assert err is None

        # Get Oauth Scope
        found_oauth_scope, resp, err = await client.get_o_auth_2_scope(
            created_auth_server.id, oauth_scope.id
        )
        assert err is not None
        assert isinstance(err, OktaAPIError)
        assert resp.get_status() == HTTPStatus.NOT_FOUND

        # DeActivate Auth server
        _, err = await client.deactivate_authorization_server(
            created_auth_server.id)
        assert err is None
        # Delete Auth server
        _, err = await client.delete_authorization_server(
            created_auth_server.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_create_get_oauth2_claim(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Auth Server
        TEST_NAME = f"{TestAuthorizationServerResource.SDK_PREFIX}_test_abDfZ"
        TEST_DESC = "Test Auth Server"
        TEST_AUDS = ["api://default"]
        auth_server_model = models.AuthorizationServer({
            "name": TEST_NAME,
            "description": TEST_DESC,
            "audiences": TEST_AUDS
        })

        created_auth_server, _, err = await \
            client.create_authorization_server(auth_server_model)
        assert err is None
        assert isinstance(created_auth_server, models.AuthorizationServer)
        assert created_auth_server.name == TEST_NAME
        assert created_auth_server.description == TEST_DESC
        assert created_auth_server.audiences == TEST_AUDS
        assert created_auth_server.audiences[0] == TEST_AUDS[0]

        # Create Oauth Claim
        CLAIM_NAME = f"{TestAuthorizationServerResource.SDK_PREFIX}_abeIz"
        CLAIM_STATUS = "INACTIVE"
        CLAIM_TYPE = "RESOURCE"
        CLAIM_VALUE_TYPE = "EXPRESSION"
        CLAIM_VALUE = "\"driving!\""
        claim_obj = models.OAuth2Claim({
            "name": CLAIM_NAME,
            "status": CLAIM_STATUS,
            "claimType": CLAIM_TYPE,
            "valueType": CLAIM_VALUE_TYPE,
            "value": CLAIM_VALUE
        })

        oauth_claim, _, err = await client.create_o_auth_2_claim(
            created_auth_server.id, claim_obj
        )
        assert err is None
        assert isinstance(oauth_claim, models.OAuth2Claim)
        assert oauth_claim.name == CLAIM_NAME

        # Get Oauth claim
        found_oauth_claim, _, err = await client.get_o_auth_2_claim(
            created_auth_server.id, oauth_claim.id
        )
        assert err is None
        assert found_oauth_claim.name == oauth_claim.name

        # Delete Oauth claim
        _, err = await client.delete_o_auth_2_claim(
            created_auth_server.id, oauth_claim.id
        )
        assert err is None

        # DeActivate Auth server
        _, err = await client.deactivate_authorization_server(
            created_auth_server.id)
        assert err is None
        # Delete Auth server
        _, err = await client.delete_authorization_server(
            created_auth_server.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_list_oauth2_claims(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Auth Server
        TEST_NAME = f"{TestAuthorizationServerResource.SDK_PREFIX}_test_abDfZ"
        TEST_DESC = "Test Auth Server"
        TEST_AUDS = ["api://default"]
        auth_server_model = models.AuthorizationServer({
            "name": TEST_NAME,
            "description": TEST_DESC,
            "audiences": TEST_AUDS
        })

        created_auth_server, _, err = await \
            client.create_authorization_server(auth_server_model)
        assert err is None
        assert isinstance(created_auth_server, models.AuthorizationServer)
        assert created_auth_server.name == TEST_NAME
        assert created_auth_server.description == TEST_DESC
        assert created_auth_server.audiences == TEST_AUDS
        assert created_auth_server.audiences[0] == TEST_AUDS[0]

        # Create Oauth Claim
        CLAIM_NAME = f"{TestAuthorizationServerResource.SDK_PREFIX}_abeIz"
        CLAIM_STATUS = "INACTIVE"
        CLAIM_TYPE = "RESOURCE"
        CLAIM_VALUE_TYPE = "EXPRESSION"
        CLAIM_VALUE = "\"driving!\""
        claim_obj = models.OAuth2Claim({
            "name": CLAIM_NAME,
            "status": CLAIM_STATUS,
            "claimType": CLAIM_TYPE,
            "valueType": CLAIM_VALUE_TYPE,
            "value": CLAIM_VALUE
        })

        oauth_claim, _, err = await client.create_o_auth_2_claim(
            created_auth_server.id, claim_obj
        )
        assert err is None
        assert isinstance(oauth_claim, models.OAuth2Claim)
        assert oauth_claim.name == CLAIM_NAME

        # List Oauth claims
        claims, _, err = await client.list_o_auth_2_claims(
            created_auth_server.id)
        assert err is None
        assert next((claim for claim in claims
                     if claim.id == oauth_claim.id))

        # Delete Oauth claim
        _, err = await client.delete_o_auth_2_claim(
            created_auth_server.id, oauth_claim.id
        )
        assert err is None

        # DeActivate Auth server
        _, err = await client.deactivate_authorization_server(
            created_auth_server.id)
        assert err is None
        # Delete Auth server
        _, err = await client.delete_authorization_server(
            created_auth_server.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_update_oauth2_claim(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Auth Server
        TEST_NAME = f"{TestAuthorizationServerResource.SDK_PREFIX}_test_abDfZ"
        TEST_DESC = "Test Auth Server"
        TEST_AUDS = ["api://default"]
        auth_server_model = models.AuthorizationServer({
            "name": TEST_NAME,
            "description": TEST_DESC,
            "audiences": TEST_AUDS
        })

        created_auth_server, _, err = await \
            client.create_authorization_server(auth_server_model)
        assert err is None
        assert isinstance(created_auth_server, models.AuthorizationServer)
        assert created_auth_server.name == TEST_NAME
        assert created_auth_server.description == TEST_DESC
        assert created_auth_server.audiences == TEST_AUDS
        assert created_auth_server.audiences[0] == TEST_AUDS[0]

        # Create Oauth Claim
        CLAIM_NAME = f"{TestAuthorizationServerResource.SDK_PREFIX}_abeIz"
        CLAIM_STATUS = "INACTIVE"
        CLAIM_TYPE = "RESOURCE"
        CLAIM_VALUE_TYPE = "EXPRESSION"
        CLAIM_VALUE = "\"driving!\""
        claim_obj = models.OAuth2Claim({
            "name": CLAIM_NAME,
            "status": CLAIM_STATUS,
            "claimType": CLAIM_TYPE,
            "valueType": CLAIM_VALUE_TYPE,
            "value": CLAIM_VALUE
        })
        UPDATED_CLAIM_NAME = f"{CLAIM_NAME}_updated"
        updated_obj = models.OAuth2Claim({
            "name": UPDATED_CLAIM_NAME,
            "status": CLAIM_STATUS,
            "claimType": CLAIM_TYPE,
            "valueType": CLAIM_VALUE_TYPE,
            "value": CLAIM_VALUE
        })

        oauth_claim, _, err = await client.create_o_auth_2_claim(
            created_auth_server.id, claim_obj
        )
        assert err is None
        assert isinstance(oauth_claim, models.OAuth2Claim)
        assert oauth_claim.name == CLAIM_NAME

        # Update claim
        updated_claim, _, err = await client.update_o_auth_2_claim(
            created_auth_server.id, oauth_claim.id, updated_obj
        )
        assert err is None
        assert isinstance(oauth_claim, models.OAuth2Claim)
        assert updated_claim.name == UPDATED_CLAIM_NAME

        # Get Oauth claim
        found_oauth_claim, _, err = await client.get_o_auth_2_claim(
            created_auth_server.id, oauth_claim.id
        )
        assert err is None
        assert found_oauth_claim.name == UPDATED_CLAIM_NAME

        # Delete Oauth claim
        _, err = await client.delete_o_auth_2_claim(
            created_auth_server.id, oauth_claim.id
        )
        assert err is None

        # DeActivate Auth server
        _, err = await client.deactivate_authorization_server(
            created_auth_server.id)
        assert err is None
        # Delete Auth server
        _, err = await client.delete_authorization_server(
            created_auth_server.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_delete_oauth2_claim(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Auth Server
        TEST_NAME = f"{TestAuthorizationServerResource.SDK_PREFIX}_test_abDfZ"
        TEST_DESC = "Test Auth Server"
        TEST_AUDS = ["api://default"]
        auth_server_model = models.AuthorizationServer({
            "name": TEST_NAME,
            "description": TEST_DESC,
            "audiences": TEST_AUDS
        })

        created_auth_server, _, err = await \
            client.create_authorization_server(auth_server_model)
        assert err is None
        assert isinstance(created_auth_server, models.AuthorizationServer)
        assert created_auth_server.name == TEST_NAME
        assert created_auth_server.description == TEST_DESC
        assert created_auth_server.audiences == TEST_AUDS
        assert created_auth_server.audiences[0] == TEST_AUDS[0]

        # Create Oauth Claim
        CLAIM_NAME = f"{TestAuthorizationServerResource.SDK_PREFIX}_abeIz"
        CLAIM_STATUS = "INACTIVE"
        CLAIM_TYPE = "RESOURCE"
        CLAIM_VALUE_TYPE = "EXPRESSION"
        CLAIM_VALUE = "\"driving!\""
        claim_obj = models.OAuth2Claim({
            "name": CLAIM_NAME,
            "status": CLAIM_STATUS,
            "claimType": CLAIM_TYPE,
            "valueType": CLAIM_VALUE_TYPE,
            "value": CLAIM_VALUE
        })

        oauth_claim, _, err = await client.create_o_auth_2_claim(
            created_auth_server.id, claim_obj
        )
        assert err is None
        assert isinstance(oauth_claim, models.OAuth2Claim)
        assert oauth_claim.name == CLAIM_NAME

        # Get Oauth claim
        found_oauth_claim, _, err = await client.get_o_auth_2_claim(
            created_auth_server.id, oauth_claim.id
        )
        assert err is None
        assert found_oauth_claim.name == oauth_claim.name

        # Delete Oauth claim
        _, err = await client.delete_o_auth_2_claim(
            created_auth_server.id, oauth_claim.id
        )
        assert err is None

        # Get Oauth claim
        found_oauth_claim, resp, err = await client.get_o_auth_2_claim(
            created_auth_server.id, oauth_claim.id
        )
        assert err is not None
        assert isinstance(err, OktaAPIError)
        assert resp.get_status() == HTTPStatus.NOT_FOUND

        # DeActivate Auth server
        _, err = await client.deactivate_authorization_server(
            created_auth_server.id)
        assert err is None
        # Delete Auth server
        _, err = await client.delete_authorization_server(
            created_auth_server.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_list_auth_server_keys(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Auth Server
        TEST_NAME = f"{TestAuthorizationServerResource.SDK_PREFIX}_test_abDfZ"
        TEST_DESC = "Test Auth Server"
        TEST_AUDS = ["api://default"]
        auth_server_model = models.AuthorizationServer({
            "name": TEST_NAME,
            "description": TEST_DESC,
            "audiences": TEST_AUDS
        })

        created_auth_server, _, err = await \
            client.create_authorization_server(auth_server_model)
        assert err is None
        assert isinstance(created_auth_server, models.AuthorizationServer)
        assert created_auth_server.name == TEST_NAME
        assert created_auth_server.description == TEST_DESC
        assert created_auth_server.audiences == TEST_AUDS
        assert created_auth_server.audiences[0] == TEST_AUDS[0]

        # List keys
        auth_server_keys, _, err = await\
            client.list_authorization_server_keys(
                created_auth_server.id
            )
        assert err is None
        assert len(auth_server_keys) > 0

        # DeActivate Auth server
        _, err = await client.deactivate_authorization_server(
            created_auth_server.id)
        assert err is None
        # Delete Auth server
        _, err = await client.delete_authorization_server(
            created_auth_server.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_rotate_auth_server_keys(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Auth Server
        TEST_NAME = f"{TestAuthorizationServerResource.SDK_PREFIX}_test_abDfZ"
        TEST_DESC = "Test Auth Server"
        TEST_AUDS = ["api://default"]
        auth_server_model = models.AuthorizationServer({
            "name": TEST_NAME,
            "description": TEST_DESC,
            "audiences": TEST_AUDS
        })

        created_auth_server, _, err = await \
            client.create_authorization_server(auth_server_model)
        assert err is None
        assert isinstance(created_auth_server, models.AuthorizationServer)
        assert created_auth_server.name == TEST_NAME
        assert created_auth_server.description == TEST_DESC
        assert created_auth_server.audiences == TEST_AUDS
        assert created_auth_server.audiences[0] == TEST_AUDS[0]

        # Make key to rotate
        rotate_key = models.JwkUse({
            "use": "sig"
        })

        # Rotate key
        keys, _, err = await client.rotate_authorization_server_keys(
            created_auth_server.id, rotate_key
        )
        assert err is None
        assert len(keys) > 0

        # DeActivate Auth server
        _, err = await client.deactivate_authorization_server(
            created_auth_server.id)
        assert err is None
        # Delete Auth server
        _, err = await client.delete_authorization_server(
            created_auth_server.id)
        assert err is None



import pytest
from tests.mocks import MockOktaClient
import okta.models as models
from okta.constants import DATETIME_FORMAT
import datetime as dt


class TestLogEventsResource:
    """
    Integration Tests for the Log Events Resource
    """
    SDK_PREFIX = "python_sdk"

    @pytest.mark.asyncio
    @pytest.mark.skip
    async def test_get_logs(self):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Bookmark Application Object
        APP_URL = "https://example.com/bookmark.htm"
        APP_LABEL = "AddBookmarkApp-GetLogs"
        app_settings_app = models.BookmarkApplicationSettingsApplication({
            "requestIntegration": False,
            "url": APP_URL
        })
        app_settings = models.BookmarkApplicationSettings({
            "app": app_settings_app
        })
        bookmark_app_obj = models.BookmarkApplication({
            "label": APP_LABEL,
            "settings": app_settings
        })

        # Create App in org
        app, _, err = await client.create_application(bookmark_app_obj)
        assert err is None
        assert isinstance(app, models.Application)
        assert isinstance(app, models.BookmarkApplication)

        logs, _, err = await client.get_logs()
        assert err is None
        assert logs is not None
        assert isinstance(logs, list)
        if logs[0]:
            assert isinstance(logs[0], models.LogEvent)

        # Deactivate & Delete created app
        _, err = await client.deactivate_application(app.id)
        assert err is None
        _, err = await client.delete_application(app.id)
        assert err is None

    @pytest.mark.asyncio
    @pytest.mark.skip
    async def test_get_logs_polling(self):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Bookmark Application Object
        APP_URL = "https://example.com/bookmark.htm"
        APP_LABEL = "AddBookmarkApp-GetLogs"
        app_settings_app = models.BookmarkApplicationSettingsApplication({
            "requestIntegration": False,
            "url": APP_URL
        })
        app_settings = models.BookmarkApplicationSettings({
            "app": app_settings_app
        })
        bookmark_app_obj = models.BookmarkApplication({
            "label": APP_LABEL,
            "settings": app_settings
        })

        # Create App in org
        app, _, err = await client.create_application(bookmark_app_obj)
        assert err is None
        assert isinstance(app, models.Application)
        assert isinstance(app, models.BookmarkApplication)

        now = dt.datetime.now(dt.timezone.utc)
        now = now.replace(microsecond=0)
        one_minute_after = now + dt.timedelta(minutes=1)

        log_query_params = {
            "sortOrder": "ASCENDING",
            "until": one_minute_after.strftime(DATETIME_FORMAT)
        }

        logs, _, err = await client.get_logs(log_query_params)
        assert err is None
        assert logs is not None
        assert isinstance(logs, list)
        if logs[0]:
            assert isinstance(logs[0], models.LogEvent)

        # Deactivate & Delete created app
        _, err = await client.deactivate_application(app.id)
        assert err is None
        _, err = await client.delete_application(app.id)
        assert err is None

    @pytest.mark.asyncio
    @pytest.mark.skip
    async def test_get_logs_bounded(self):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Bookmark Application Object
        APP_URL = "https://example.com/bookmark.htm"
        APP_LABEL = "AddBookmarkApp-GetLogs"
        app_settings_app = models.BookmarkApplicationSettingsApplication({
            "requestIntegration": False,
            "url": APP_URL
        })
        app_settings = models.BookmarkApplicationSettings({
            "app": app_settings_app
        })
        bookmark_app_obj = models.BookmarkApplication({
            "label": APP_LABEL,
            "settings": app_settings
        })

        # Create App in org
        app, _, err = await client.create_application(bookmark_app_obj)
        assert err is None
        assert isinstance(app, models.Application)
        assert isinstance(app, models.BookmarkApplication)

        # Retrieve logs
        now = dt.datetime.now(dt.timezone.utc)
        now = now.replace(microsecond=0)
        one_hour_before = now - dt.timedelta(hours=1)

        log_query_params = {
            "since": one_hour_before.strftime(DATETIME_FORMAT),
            "until": now.strftime(DATETIME_FORMAT)
        }
        logs, _, err = await client.get_logs(log_query_params)
        assert err is None
        assert logs is not None
        assert isinstance(logs, list)
        if logs[0]:
            assert isinstance(logs[0], models.LogEvent)

        # Deactivate & Delete created app
        _, err = await client.deactivate_application(app.id)
        assert err is None
        _, err = await client.delete_application(app.id)
        assert err is None



"""
Copyright 2020 - Present Okta, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# AUTO-GENERATED! DO NOT EDIT FILE DIRECTLY
# SEE CONTRIBUTOR DOCUMENTATION

from urllib.parse import urlencode
from okta.models.feature\
    import Feature
from okta.utils import format_url
from okta.api_client import APIClient


class FeatureClient(APIClient):
    """
    A Client object for the Feature resource.
    """

    def __init__(self):
        self._base_url = ""

    async def list_features(
            self
    ):
        """
        Args:
        Returns:
            list: Collection of Feature instances.
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/features
            """)

        body = {}
        headers = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, Feature)

        if error:
            return (None, response, error)

        try:
            result = []
            for item in response.get_body():
                result.append(Feature(
                    self.form_response_body(item)
                    ))
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def get_feature(
            self, featureId
    ):
        """
        Args:
            feature_id {str}
        Returns:
            Feature
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/features/{featureId}
            """)

        body = {}
        headers = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, Feature)

        if error:
            return (None, response, error)

        try:
            result = Feature(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def list_feature_dependencies(
            self, featureId
    ):
        """
        Args:
            feature_id {str}
        Returns:
            list: Collection of Feature instances.
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/features/{featureId}/dependencies
            """)

        body = {}
        headers = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, Feature)

        if error:
            return (None, response, error)

        try:
            result = []
            for item in response.get_body():
                result.append(Feature(
                    self.form_response_body(item)
                    ))
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def list_feature_dependents(
            self, featureId
    ):
        """
        Args:
            feature_id {str}
        Returns:
            list: Collection of Feature instances.
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/features/{featureId}/dependents
            """)

        body = {}
        headers = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, Feature)

        if error:
            return (None, response, error)

        try:
            result = []
            for item in response.get_body():
                result.append(Feature(
                    self.form_response_body(item)
                    ))
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def update_feature_lifecycle(
            self, featureId, lifecycle, query_params={}
    ):
        """
        Args:
            feature_id {str}
            lifecycle {str}
            query_params {dict}: Map of query parameters for request
            [query_params.mode] {str}
        Returns:
            Feature
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/features/{featureId}/{lifecycle}
            """)
        if query_params:
            encoded_query_params = urlencode(query_params)
            api_url += f"/?{encoded_query_params}"

        body = {}
        headers = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, Feature)

        if error:
            return (None, response, error)

        try:
            result = Feature(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)



"""
Copyright 2020 - Present Okta, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# AUTO-GENERATED! DO NOT EDIT FILE DIRECTLY
# SEE CONTRIBUTOR DOCUMENTATION

import logging

from okta.config.config_setter import ConfigSetter
from okta.config.config_validator import ConfigValidator
from okta.request_executor import RequestExecutor
from okta.cache.no_op_cache import NoOpCache
from okta.cache.okta_cache import OktaCache
from okta.logger import setup_logging
from okta.resource_clients.application_client\
    import ApplicationClient
from okta.resource_clients.authorization_server_client\
    import AuthorizationServerClient
from okta.resource_clients.event_hook_client\
    import EventHookClient
from okta.resource_clients.feature_client\
    import FeatureClient
from okta.resource_clients.group_client\
    import GroupClient
from okta.resource_clients.identity_provider_client\
    import IdentityProviderClient
from okta.resource_clients.inline_hook_client\
    import InlineHookClient
from okta.resource_clients.log_event_client\
    import LogEventClient
from okta.resource_clients.linked_object_client\
    import LinkedObjectClient
from okta.resource_clients.user_type_client\
    import UserTypeClient
from okta.resource_clients.policy_client\
    import PolicyClient
from okta.resource_clients.session_client\
    import SessionClient
from okta.resource_clients.sms_template_client\
    import SmsTemplateClient
from okta.resource_clients.trusted_origin_client\
    import TrustedOriginClient
from okta.resource_clients.user_client\
    import UserClient
from okta.resource_clients.user_factor_client\
    import UserFactorClient


class Client(
    ApplicationClient,
    AuthorizationServerClient,
    EventHookClient,
    FeatureClient,
    GroupClient,
    IdentityProviderClient,
    InlineHookClient,
    LogEventClient,
    LinkedObjectClient,
    UserTypeClient,
    PolicyClient,
    SessionClient,
    SmsTemplateClient,
    TrustedOriginClient,
    UserClient,
    UserFactorClient
):
    """An Okta client object"""

    def __init__(self, user_config: dict = {}):
        super()
        # Load configuration
        client_config_setter = ConfigSetter()
        client_config_setter._apply_config({'client': user_config})
        self._config = client_config_setter.get_config()
        # Prune configuration to remove unnecesary fields
        self._config = client_config_setter._prune_config(self._config)
        # Validate configuration
        ConfigValidator(self._config)

        # set client variables since validation passes
        self._authorization_mode = self._config["client"]["authorizationMode"]
        self._base_url = self._config["client"]["orgUrl"]
        self._api_token = self._config["client"].get("token", None)
        self._client_id = None
        self._scopes = None
        self._private_key = None

        # Determine which cache to use
        cache = NoOpCache()
        if self._config["client"]["cache"]["enabled"] is True:
            if user_config.get("cacheManager", None) is None:
                time_to_idle = self._config["client"]["cache"]["defaultTti"]
                time_to_live = self._config["client"]["cache"]["defaultTtl"]
                cache = OktaCache(
                    time_to_live,
                    time_to_idle
                )
            else:
                cache = user_config.get("cacheManager")

        # Determine request executor to use
        self._request_executor = \
            user_config.get("requestExecutor", RequestExecutor)(
                self._config,
                cache,
                user_config.get("httpClient", None))

        # set private key variables
        if self._authorization_mode == 'PrivateKey':
            self._client_id = self._config["client"]["clientId"]
            self._scopes = self._config["client"]["scopes"]
            self._private_key = self._config["client"]["privateKey"]

        setup_logging(log_level=self._config["client"]["logging"]["logLevel"])
        # Check if logging should be enabled
        if self._config["client"]["logging"]["enabled"] is True:
            logger = logging.getLogger('okta-sdk-python')
            logger.disabled = False
    """
    Getters
    """

    def get_config(self):
        return self._config

    def get_scopes(self):
        return self._scopes

    def get_base_url(self):
        return self._base_url

    def get_request_executor(self):
        return self._request_executor

    """
    Misc
    """
    def set_custom_headers(self, headers):
        self._request_executor.set_custom_headers(headers)

    def clear_custom_headers(self):
        self._request_executor.clear_custom_headers()

    def get_custom_headers(self):
        return self._request_executor.get_custom_headers()

    def get_default_headers(self):
        return self._request_executor.get_default_headers()
